# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import logging
import mimetypes
import os
import re
import threading
from typing import Optional

from typing_extensions import override

from ..agents.invocation_context import InvocationContext
from .base_code_executor import BaseCodeExecutor
from .code_execution_utils import CodeExecutionInput
from .code_execution_utils import CodeExecutionResult
from .code_execution_utils import File

logger = logging.getLogger('google_adk.' + __name__)


class AgentEngineSandboxCodeExecutor(BaseCodeExecutor):
  """A code executor that uses Agent Engine Code Execution Sandbox to execute code.

  Attributes:
    sandbox_resource_name: If set, load the existing resource name of the code
      interpreter extension instead of creating a new one. Format:
      projects/123/locations/us-central1/reasoningEngines/456/sandboxEnvironments/789
    agent_engine_resource_name: The resource name of the agent engine to use
      to create the code execution sandbox. Format:
      projects/123/locations/us-central1/reasoningEngines/456
  """

  sandbox_resource_name: str = None

  agent_engine_resource_name: str = None
  _agent_engine_creation_lock: Optional[threading.Lock] = None

  def __init__(
      self,
      sandbox_resource_name: Optional[str] = None,
      agent_engine_resource_name: Optional[str] = None,
      **data,
  ):
    """Initializes the AgentEngineSandboxCodeExecutor.

    Args:
      sandbox_resource_name: If set, load the existing resource name of code
        execution sandbox, if not set, create a new one. Format:
        projects/123/locations/us-central1/reasoningEngines/456/
        sandboxEnvironments/789
      agent_engine_resource_name: The resource name of the agent engine to use
        to create the code execution sandbox. If not set, a new Agent Engine
        will be created automatically. Format:
        projects/123/locations/us-central1/reasoningEngines/456, when both
        sandbox_resource_name and agent_engine_resource_name are set,
        agent_engine_resource_name will be ignored.
      **data: Additional keyword arguments to be passed to the base class.
    """
    super().__init__(**data)
    self._agent_engine_creation_lock = threading.Lock()
    sandbox_resource_name_pattern = r'^projects/([a-zA-Z0-9-_]+)/locations/([a-zA-Z0-9-_]+)/reasoningEngines/(\d+)/sandboxEnvironments/(\d+)$'
    agent_engine_resource_name_pattern = r'^projects/([a-zA-Z0-9-_]+)/locations/([a-zA-Z0-9-_]+)/reasoningEngines/(\d+)$'

    # Case 1: sandbox_resource_name is provided.
    if sandbox_resource_name is not None:
      self._project_id, self._location = (
          self._get_project_id_and_location_from_resource_name(
              sandbox_resource_name, sandbox_resource_name_pattern
          )
      )
      self.sandbox_resource_name = sandbox_resource_name

    # Case 2: Agent Engine resource name is not provided.
    elif agent_engine_resource_name is None:
      # The Agent Engine will be auto-created lazily within execute_code().
      self._project_id = os.environ.get('GOOGLE_CLOUD_PROJECT')
      self._location = os.environ.get('GOOGLE_CLOUD_LOCATION', 'us-central1')
      self.agent_engine_resource_name = None

    # Case 3: Use the provided agent_engine_resource_name.
    else:
      self._project_id, self._location = (
          self._get_project_id_and_location_from_resource_name(
              agent_engine_resource_name,
              agent_engine_resource_name_pattern,
          )
      )
      self.agent_engine_resource_name = agent_engine_resource_name

  @override
  def execute_code(
      self,
      invocation_context: InvocationContext,
      code_execution_input: CodeExecutionInput,
  ) -> CodeExecutionResult:
    if (
        self.sandbox_resource_name is None
        and self.agent_engine_resource_name is None
    ):
      with self._agent_engine_creation_lock:
        if self.agent_engine_resource_name is None:
          logger.info(
              'No Agent Engine resource name provided. Creating a new one...'
          )
          try:
            # Create a default Agent Engine.
            created_engine = self._get_api_client().agent_engines.create()
            self.agent_engine_resource_name = created_engine.api_resource.name
            logger.info(
                'Created Agent Engine: %s', self.agent_engine_resource_name
            )
          except Exception as e:
            logger.error('Failed to auto-create Agent Engine: %s', e)
            raise
    # default to the sandbox resource name if set.
    sandbox_name = self.sandbox_resource_name
    if self.sandbox_resource_name is None:
      from google.api_core import exceptions
      from vertexai import types

      # use sandbox name stored in session if available.
      sandbox_name = invocation_context.session.state.get('sandbox_name', None)
      create_new_sandbox = False
      if sandbox_name is None:
        create_new_sandbox = True
      else:
        # Check if the sandbox is still running OR already expired due to ttl.
        try:
          sandbox = self._get_api_client().agent_engines.sandboxes.get(
              name=sandbox_name
          )
          if sandbox is None or sandbox.state != 'STATE_RUNNING':
            create_new_sandbox = True
        except exceptions.NotFound:
          create_new_sandbox = True

      if create_new_sandbox:
        # Create a new sandbox and assign it to sandbox_name.
        operation = self._get_api_client().agent_engines.sandboxes.create(
            spec={'code_execution_environment': {}},
            name=self.agent_engine_resource_name,
            config=types.CreateAgentEngineSandboxConfig(
                # VertexAiSessionService has a default TTL of 1 year, so we set
                # the sandbox TTL to 1 year as well. For the current code
                # execution sandbox, if it hasn't been used for 14 days, the
                # state will be lost.
                display_name='default_sandbox',
                ttl='31536000s',
            ),
        )
        sandbox_name = operation.response.name
        invocation_context.session.state['sandbox_name'] = sandbox_name

    # Execute the code.
    input_data = {
        'code': code_execution_input.code,
    }
    if code_execution_input.input_files:
      input_data['files'] = [
          {
              'name': f.name,
              'contents': f.content,
              'mimeType': f.mime_type,
          }
          for f in code_execution_input.input_files
      ]

    code_execution_response = (
        self._get_api_client().agent_engines.sandboxes.execute_code(
            name=sandbox_name,
            input_data=input_data,
        )
    )
    logger.debug('Executed code:\n```\n%s\n```', code_execution_input.code)
    saved_files = []
    stdout = ''
    stderr = ''
    for output in code_execution_response.outputs:
      if output.mime_type == 'application/json' and (
          output.metadata is None
          or output.metadata.attributes is None
          or 'file_name' not in output.metadata.attributes
      ):
        json_output_data = json.loads(output.data.decode('utf-8'))
        stdout = json_output_data.get('msg_out', '')
        stderr = json_output_data.get('msg_err', '')
      else:
        file_name = ''
        if (
            output.metadata is not None
            and output.metadata.attributes is not None
        ):
          file_name = output.metadata.attributes.get('file_name', b'').decode(
              'utf-8'
          )
        mime_type = output.mime_type
        if not mime_type:
          mime_type, _ = mimetypes.guess_type(file_name)
        saved_files.append(
            File(
                name=file_name,
                content=output.data,
                mime_type=mime_type,
            )
        )

    # Collect the final result.
    return CodeExecutionResult(
        stdout=stdout,
        stderr=stderr,
        output_files=saved_files,
    )

  def _get_api_client(self):
    """Instantiates an API client for the given project and location.

    It needs to be instantiated inside each request so that the event loop
    management can be properly propagated.

    Returns:
      An API client for the given project and location.
    """
    import vertexai

    return vertexai.Client(project=self._project_id, location=self._location)

  def _get_project_id_and_location_from_resource_name(
      self, resource_name: str, pattern: str
  ) -> tuple[str, str]:
    """Extracts the project ID and location from the resource name."""
    match = re.fullmatch(pattern, resource_name)

    if not match:
      raise ValueError(f'resource name {resource_name} is not valid.')

    return match.groups()[0], match.groups()[1]



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any
from typing import Optional

from google.genai import types
from pydantic import alias_generators
from pydantic import BaseModel
from pydantic import ConfigDict

from .cache_metadata import CacheMetadata


class LlmResponse(BaseModel):
  """LLM response class that provides the first candidate response from the

  model if available. Otherwise, returns error code and message.

  Attributes:
    content: The content of the response.
    grounding_metadata: The grounding metadata of the response.
    partial: Indicates whether the text content is part of an unfinished text
      stream. Only used for streaming mode and when the content is plain text.
    turn_complete: Indicates whether the response from the model is complete.
      Only used for streaming mode.
    error_code: Error code if the response is an error. Code varies by model.
    error_message: Error message if the response is an error.
    interrupted: Flag indicating that LLM was interrupted when generating the
      content. Usually it's due to user interruption during a bidi streaming.
    custom_metadata: The custom metadata of the LlmResponse.
    input_transcription: Audio transcription of user input.
    output_transcription: Audio transcription of model output.
    avg_logprobs: Average log probability of the generated tokens.
    logprobs_result: Detailed log probabilities for chosen and top candidate tokens.
  """

  model_config = ConfigDict(
      extra='forbid',
      alias_generator=alias_generators.to_camel,
      populate_by_name=True,
  )
  """The pydantic model config."""

  model_version: Optional[str] = None
  """Output only. The model version used to generate the response."""

  content: Optional[types.Content] = None
  """The generative content of the response.

  This should only contain content from the user or the model, and not any
  framework or system-generated data.
  """

  grounding_metadata: Optional[types.GroundingMetadata] = None
  """The grounding metadata of the response."""

  partial: Optional[bool] = None
  """Indicates whether the text content is part of an unfinished text stream.

  Only used for streaming mode and when the content is plain text.
  """

  turn_complete: Optional[bool] = None
  """Indicates whether the response from the model is complete.

  Only used for streaming mode.
  """

  finish_reason: Optional[types.FinishReason] = None
  """The finish reason of the response."""

  error_code: Optional[str] = None
  """Error code if the response is an error. Code varies by model."""

  error_message: Optional[str] = None
  """Error message if the response is an error."""

  interrupted: Optional[bool] = None
  """Flag indicating that LLM was interrupted when generating the content.
  Usually it's due to user interruption during a bidi streaming.
  """

  custom_metadata: Optional[dict[str, Any]] = None
  """The custom metadata of the LlmResponse.

  An optional key-value pair to label an LlmResponse.

  NOTE: the entire dict must be JSON serializable.
  """

  usage_metadata: Optional[types.GenerateContentResponseUsageMetadata] = None
  """The usage metadata of the LlmResponse"""

  live_session_resumption_update: Optional[
      types.LiveServerSessionResumptionUpdate
  ] = None
  """The session resumption update of the LlmResponse"""

  input_transcription: Optional[types.Transcription] = None
  """Audio transcription of user input."""

  output_transcription: Optional[types.Transcription] = None
  """Audio transcription of model output."""

  avg_logprobs: Optional[float] = None
  """Average log probability of the generated tokens."""

  logprobs_result: Optional[types.LogprobsResult] = None
  """Detailed log probabilities for chosen and top candidate tokens."""

  cache_metadata: Optional[CacheMetadata] = None
  """Context cache metadata if caching was used for this response.

  Contains cache identification, usage tracking, and lifecycle information.
  This field is automatically populated when context caching is enabled.
  """

  citation_metadata: Optional[types.CitationMetadata] = None
  """Citation metadata for the response.

  This field is automatically populated when citation is enabled.
  """

  interaction_id: Optional[str] = None
  """The interaction ID from the interactions API.

  This field is populated when using the interactions API for model invocation.
  It can be used to identify and chain interactions for stateful conversations.
  """

  @staticmethod
  def create(
      generate_content_response: types.GenerateContentResponse,
  ) -> LlmResponse:
    """Creates an LlmResponse from a GenerateContentResponse.

    Args:
      generate_content_response: The GenerateContentResponse to create the
        LlmResponse from.

    Returns:
      The LlmResponse.
    """
    usage_metadata = generate_content_response.usage_metadata
    if generate_content_response.candidates:
      candidate = generate_content_response.candidates[0]
      if (
          candidate.content and candidate.content.parts
      ) or candidate.finish_reason == types.FinishReason.STOP:
        return LlmResponse(
            content=candidate.content,
            grounding_metadata=candidate.grounding_metadata,
            usage_metadata=usage_metadata,
            finish_reason=candidate.finish_reason,
            citation_metadata=candidate.citation_metadata,
            avg_logprobs=candidate.avg_logprobs,
            logprobs_result=candidate.logprobs_result,
            model_version=generate_content_response.model_version,
        )
      else:
        return LlmResponse(
            error_code=candidate.finish_reason,
            error_message=candidate.finish_message,
            citation_metadata=candidate.citation_metadata,
            usage_metadata=usage_metadata,
            finish_reason=candidate.finish_reason,
            avg_logprobs=candidate.avg_logprobs,
            logprobs_result=candidate.logprobs_result,
            model_version=generate_content_response.model_version,
        )
    else:
      if generate_content_response.prompt_feedback:
        prompt_feedback = generate_content_response.prompt_feedback
        return LlmResponse(
            error_code=prompt_feedback.block_reason,
            error_message=prompt_feedback.block_reason_message,
            usage_metadata=usage_metadata,
            model_version=generate_content_response.model_version,
        )
      else:
        return LlmResponse(
            error_code='UNKNOWN_ERROR',
            error_message='Unknown error.',
            usage_metadata=usage_metadata,
            model_version=generate_content_response.model_version,
        )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging

from .base_code_executor import BaseCodeExecutor
from .built_in_code_executor import BuiltInCodeExecutor
from .code_executor_context import CodeExecutorContext
from .unsafe_local_code_executor import UnsafeLocalCodeExecutor

logger = logging.getLogger('google_adk.' + __name__)

__all__ = [
    'BaseCodeExecutor',
    'BuiltInCodeExecutor',
    'CodeExecutorContext',
    'UnsafeLocalCodeExecutor',
    'VertexAiCodeExecutor',
    'ContainerCodeExecutor',
    'GkeCodeExecutor',
    'AgentEngineSandboxCodeExecutor',
]


def __getattr__(name: str):
  if name == 'VertexAiCodeExecutor':
    try:
      from .vertex_ai_code_executor import VertexAiCodeExecutor

      return VertexAiCodeExecutor
    except ImportError as e:
      raise ImportError(
          'VertexAiCodeExecutor requires additional dependencies. '
          'Please install with: pip install "google-adk[extensions]"'
      ) from e
  elif name == 'ContainerCodeExecutor':
    try:
      from .container_code_executor import ContainerCodeExecutor

      return ContainerCodeExecutor
    except ImportError as e:
      raise ImportError(
          'ContainerCodeExecutor requires additional dependencies. '
          'Please install with: pip install "google-adk[extensions]"'
      ) from e
  elif name == 'GkeCodeExecutor':
    try:
      from .gke_code_executor import GkeCodeExecutor

      return GkeCodeExecutor
    except ImportError as e:
      raise ImportError(
          'GkeCodeExecutor requires additional dependencies. '
          'Please install with: pip install "google-adk[extensions]"'
      ) from e
  elif name == 'AgentEngineSandboxCodeExecutor':
    try:
      from .agent_engine_sandbox_code_executor import AgentEngineSandboxCodeExecutor

      return AgentEngineSandboxCodeExecutor
    except ImportError as e:
      raise ImportError(
          'AgentEngineSandboxCodeExecutor requires additional dependencies. '
          'Please install with: pip install "google-adk[extensions]"'
      ) from e
  raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Optional

from pydantic import Field
from pydantic import field_validator

from .common import EvalBaseModel
from .simulation.pre_built_personas import get_default_persona_registry
from .simulation.user_simulator_personas import UserPersona


class ConversationScenario(EvalBaseModel):
  """Scenario for a conversation between a simulated user and the Agent under test."""

  starting_prompt: str
  """Starting prompt for the conversation.

  This prompt acts as the fixed first user message that is given to the Agent.
  Any subsequent user messages are obtained by the system that is simulating the
  user.
  """

  conversation_plan: str
  """A plan that user simulation system needs to follow as it plays out the conversation.

  Example:
  For a Travel Agent that has tools that let it book a flight and car, a sample
  starting prompt could be:

  `I need to book a flight.`

  A conversation plan could look like:

  First, you want to book a one-way flight from SFO to LAX for next Tuesday.
  You prefer a morning flight and your budget is under $150. If the agent finds
  a valid flight, confirm the booking. Once confirmed, your next goal is to rent
  a standard-size car for three days from the airport. Once both tasks are done,
  your overall goal is complete.
  """

  user_persona: Optional[UserPersona] = Field(default=None)
  """User persona that the user simulator should adopt. If a persona id is specified instead, we will try to use one of our default personas."""

  @field_validator("user_persona", mode="before")
  @classmethod
  def validate_user_persona(
      cls, value: Optional[UserPersona | str]
  ) -> Optional[UserPersona]:
    if value is not None and isinstance(value, str):
      return get_default_persona_registry().get_persona(value)
    return value


class ConversationScenarios(EvalBaseModel):
  """A simple container for the list of ConversationScenario.

  Mainly serves the purpose of helping with serialization and deserialization.
  """

  scenarios: list[ConversationScenario] = Field(
      default_factory=list, description="""A list of ConversationScenario."""
  )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Iterator
from contextlib import contextmanager
import contextvars
import os
import sys
from typing import List

from .. import version

_ADK_LABEL = "google-adk"
_LANGUAGE_LABEL = "gl-python"
_AGENT_ENGINE_TELEMETRY_TAG = "remote_reasoning_engine"
_AGENT_ENGINE_TELEMETRY_ENV_VARIABLE_NAME = "GOOGLE_CLOUD_AGENT_ENGINE_ID"


EVAL_CLIENT_LABEL = f"google-adk-eval/{version.__version__}"
"""Label used to denote calls emerging to external system as a part of Evals."""

# The ContextVar holds client label collected for the current request.
_LABEL_CONTEXT: contextvars.ContextVar[str | None] = contextvars.ContextVar(
    "_LABEL_CONTEXT", default=None
)


def _get_default_labels() -> List[str]:
  """Returns a list of labels that are always added."""
  framework_label = f"{_ADK_LABEL}/{version.__version__}"

  if os.environ.get(_AGENT_ENGINE_TELEMETRY_ENV_VARIABLE_NAME):
    framework_label = f"{framework_label}+{_AGENT_ENGINE_TELEMETRY_TAG}"

  language_label = f"{_LANGUAGE_LABEL}/" + sys.version.split()[0]
  return [framework_label, language_label]


@contextmanager
def client_label_context(client_label: str) -> Iterator[None]:
  """Runs the operation within the context of the given client label."""
  current_client_label = _LABEL_CONTEXT.get()

  if current_client_label is not None:
    raise ValueError(
        "Client label already exists. You can only add one client label."
    )

  token = _LABEL_CONTEXT.set(client_label)

  try:
    yield
  finally:
    # Restore the previous state of the context variable
    _LABEL_CONTEXT.reset(token)


def get_client_labels() -> List[str]:
  """Returns the current list of client labels that can be added to HTTP Headers."""
  labels = _get_default_labels()
  current_client_label = _LABEL_CONTEXT.get()

  if current_client_label:
    labels.append(current_client_label)

  return labels



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Annotated
from typing import Any
from typing import Union

from pydantic import Discriminator
from pydantic import RootModel
from pydantic import Tag

from ..features import experimental
from ..features import FeatureName
from .base_agent_config import BaseAgentConfig
from .llm_agent_config import LlmAgentConfig
from .loop_agent_config import LoopAgentConfig
from .parallel_agent_config import ParallelAgentConfig
from .sequential_agent_config import SequentialAgentConfig

_ADK_AGENT_CLASSES: set[str] = {
    "LlmAgent",
    "LoopAgent",
    "ParallelAgent",
    "SequentialAgent",
}


def agent_config_discriminator(v: Any) -> str:
  """Discriminator function that returns the tag name for Pydantic."""
  if isinstance(v, dict):
    agent_class: str = v.get("agent_class", "LlmAgent")

    # Look up the agent_class in our dynamically built mapping
    if agent_class in _ADK_AGENT_CLASSES:
      return agent_class

    # For non ADK agent classes, use BaseAgent to handle it.
    return "BaseAgent"

  raise ValueError(f"Invalid agent config: {v}")


# A discriminated union of all possible agent configurations.
ConfigsUnion = Annotated[
    Union[
        Annotated[LlmAgentConfig, Tag("LlmAgent")],
        Annotated[LoopAgentConfig, Tag("LoopAgent")],
        Annotated[ParallelAgentConfig, Tag("ParallelAgent")],
        Annotated[SequentialAgentConfig, Tag("SequentialAgent")],
        Annotated[BaseAgentConfig, Tag("BaseAgent")],
    ],
    Discriminator(agent_config_discriminator),
]


# Use a RootModel to represent the agent directly at the top level.
# The `discriminator` is applied to the union within the RootModel.
@experimental(FeatureName.AGENT_CONFIG)
class AgentConfig(RootModel[ConfigsUnion]):
  """The config for the YAML schema to create an agent."""



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import abc
from enum import Enum
from typing import Optional
from typing import Union

from google.genai import types as genai_types
from pydantic import alias_generators
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import field_validator
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import TypeAlias

from .common import EvalBaseModel
from .eval_case import Invocation
from .eval_rubrics import Rubric
from .eval_rubrics import RubricScore


class EvalStatus(Enum):
  PASSED = 1
  FAILED = 2
  NOT_EVALUATED = 3


class PrebuiltMetrics(Enum):
  TOOL_TRAJECTORY_AVG_SCORE = "tool_trajectory_avg_score"

  RESPONSE_EVALUATION_SCORE = "response_evaluation_score"

  RESPONSE_MATCH_SCORE = "response_match_score"

  SAFETY_V1 = "safety_v1"

  FINAL_RESPONSE_MATCH_V2 = "final_response_match_v2"

  RUBRIC_BASED_FINAL_RESPONSE_QUALITY_V1 = (
      "rubric_based_final_response_quality_v1"
  )

  HALLUCINATIONS_V1 = "hallucinations_v1"

  RUBRIC_BASED_TOOL_USE_QUALITY_V1 = "rubric_based_tool_use_quality_v1"

  PER_TURN_USER_SIMULATOR_QUALITY_V1 = "per_turn_user_simulator_quality_v1"

  MULTI_TURN_TASK_SUCCESS_V1 = "multi_turn_task_success_v1"

  MULTI_TURN_TRAJECTORY_QUALITY_V1 = "multi_turn_trajectory_quality_v1"

  MULTI_TURN_TOOL_USE_QUALITY_V1 = "multi_turn_tool_use_quality_v1"


MetricName: TypeAlias = Union[str, PrebuiltMetrics]
Threshold: TypeAlias = float


class JudgeModelOptions(EvalBaseModel):
  """Options for an eval metric's judge model."""

  judge_model: str = Field(
      default="gemini-2.5-flash",
      description=(
          "The judge model to use for evaluation. It can be a model name."
      ),
  )

  judge_model_config: SkipJsonSchema[
      Optional[genai_types.GenerateContentConfig]
  ] = Field(
      default=None,
      description="The configuration for the judge model.",
  )

  num_samples: int = Field(
      default=5,
      description=(
          "The number of times to sample the model for each invocation"
          " evaluation. Given that models tend to have certain degree of"
          " unreliability to them, we repeatedly sample them with the same"
          " data. These repeated invocation are them aggregated using some"
          " strategy. From experimentation, we have found 5 to be a good"
          " default."
      ),
  )


class BaseCriterion(BaseModel):
  """Base criterion to use for an Eval Metric."""

  model_config = ConfigDict(
      alias_generator=alias_generators.to_camel,
      populate_by_name=True,
      extra="allow",
  )

  threshold: Threshold = Field(
      description="The threshold to be used by the metric.",
  )


class LlmAsAJudgeCriterion(BaseCriterion):
  """Criterion when using LLM-As-A-Judge metric."""

  judge_model_options: JudgeModelOptions = Field(
      default_factory=JudgeModelOptions,
      description="Options for the judge model.",
  )


class RubricsBasedCriterion(BaseCriterion):
  """Criterion when using a rubric based metric."""

  judge_model_options: JudgeModelOptions = Field(
      default_factory=JudgeModelOptions,
      description="Options for the judge model.",
  )

  rubrics: list[Rubric] = Field(
      default_factory=list,
      description=(
          "Rubrics to be used by Metric. Not all metrics rely on rubrics, but"
          " metrics like `rubric_based_final_response_quality_v1` do. Metrics"
          " that don't use Rubrics, will just ignore this field, if specified."
          " Metrics that do use rubrics will raise an exception, if they are"
          " not specified."
      ),
  )


class HallucinationsCriterion(BaseCriterion):
  """Criterion to use when evaluating agents response for hallucinations."""

  judge_model_options: JudgeModelOptions = Field(
      default_factory=JudgeModelOptions,
      description="Options for the judge model.",
  )

  evaluate_intermediate_nl_responses: bool = Field(
      default=False,
      description=(
          "Whether any intermediate NL responses should be evaluated"
          " for hallucinations or not. By default, the metric only evaluates"
          " final response from the Agent for hallucinations."
      ),
  )


class ToolTrajectoryCriterion(BaseCriterion):
  """Criterion to use when evaluating agent's tool trajectories with a reference one."""

  class MatchType(Enum):
    """The type of Match between actual and expected tool call trajectories."""

    EXACT = 0
    """Requires a perfect match between the actual and expected tool calls."""

    IN_ORDER = 1
    """Requires the actual tool calls to be in the same order as expected tools,
    with allowance for extra tool calls to have happened.

    This criteria is useful in assuring if certain key actions/tool calls
    occur and in certain order, leaving some scope for other tools calls to
    happen as well.

    Example 1: Set of actual vs expected tool calls that satisfies the criteria:

      Expected tools calls: [T1, T2, T3]
      Actual tool calls: [T1, T1.1, T2, T2.1, T2.2, T3, T3.1]

      This satisfies, as the tools T1, T2 and T3 happened in the "Actual" and in
      the same order.

    Example 2: Set of actual vs expected tool calls that don't satisfy the
    criteria:

      Expected tools calls: [T1, T2, T3, T4]
      Actual tool calls: [T1, T1.1, T2, T2.1, T2.2, T3, T3.1]

      While the tool calls T1, T2 and T3 happened in the "Actual" and in
      the same order as "Expected", but the tool calls T4 is missing.
    """

    ANY_ORDER = 2
    """Requires the actual tool calls to be in the any order as expected tools,
    with allowance for extra tool calls to have happened.

    This criteria is helpful for cases where multiple tool calls about the same
    concept occur, like your agent issues 5 search queries. You don't really
    care the order in which the search queries are issues, till they occur.

    Example 1: Set of actual vs expected tool calls that satisfies the criteria:

      Expected tools calls: [T1, T2, T3]
      Actual tool calls: [T2, T2.1, T1, T1.1, T1.2, T3, T3.1]

      This satisfies, as the tools T1, T2 and T3 happened in the "Actual" and
      are also present in expected. Note that the order is different.

    Example 2: Set of actual vs expected tool calls that don't satisfy the
    criteria:

      Expected tools calls: [T1, T2, T3, T4]
      Actual tool calls: [T1, T1.1, T2, T2.1, T2.2, T3, T3.1]

      While the tool calls T1, T2 and T3 happened in the "Actual" and in
      the same order as "Expected", but the tool calls T4 is missing.
    """

  match_type: MatchType = Field(
      default=MatchType.EXACT,
      description=(
          "The type of Match between actual and expected tool call"
          " trajectories."
      ),
  )

  @field_validator("match_type", mode="before")
  @classmethod
  def _coerce_match_type(cls, value: object) -> object:
    if isinstance(value, cls.MatchType):
      return value
    if isinstance(value, str):
      normalized = value.strip().upper().replace("-", "_").replace(" ", "_")
      if normalized in cls.MatchType.__members__:
        return cls.MatchType[normalized]
    return value


class LlmBackedUserSimulatorCriterion(LlmAsAJudgeCriterion):
  """Criterion for LLM-backed User Simulator Evaluators."""

  stop_signal: str = Field(
      default="</finished>",
      description=(
          "Stop signal to validate the successful completion of a conversation."
          " For optimal performance, this should match the one in the User"
          " Simulator."
      ),
  )


class EvalMetric(EvalBaseModel):
  """A metric used to evaluate a particular aspect of an eval case."""

  metric_name: str = Field(
      description="The name of the metric.",
  )

  threshold: Optional[float] = Field(
      default=None,
      description=(
          "This field will be deprecated soon. Please use `criterion` instead."
          " A threshold value. Each metric decides how to interpret this"
          " threshold."
      ),
  )

  criterion: Optional[BaseCriterion] = Field(
      default=None, description="""Evaluation criterion used by the metric."""
  )

  custom_function_path: Optional[str] = Field(
      default=None,
      description="""Path to custom function, if this is a custom metric.""",
  )


class EvalMetricResultDetails(EvalBaseModel):
  rubric_scores: Optional[list[RubricScore]] = Field(
      default=None,
      description=(
          "The scores obtained after applying the rubrics to the Agent's"
          " response."
      ),
  )


class EvalMetricResult(EvalMetric):
  """The actual computed score/value of a particular EvalMetric."""

  score: Optional[float] = Field(
      default=None,
      description=(
          "Score obtained after evaluating the metric. Optional, as evaluation"
          " might not have happened."
      ),
  )

  eval_status: EvalStatus = Field(description="The status of this evaluation.")

  details: EvalMetricResultDetails = Field(
      default_factory=EvalMetricResultDetails, description=""""""
  )


class EvalMetricResultPerInvocation(EvalBaseModel):
  """Eval metric results per invocation."""

  actual_invocation: Invocation = Field(
      description=(
          "The actual invocation, usually obtained by inferencing the agent."
      )
  )

  expected_invocation: Optional[Invocation] = Field(
      default=None,
      description=(
          "The expected invocation, usually the reference or golden invocation."
      ),
  )

  eval_metric_results: list[EvalMetricResult] = Field(
      default=[],
      description="Eval results for each applicable metric.",
  )


class Interval(EvalBaseModel):
  """Represents a range of numeric values, e.g. [0 ,1] or (2,3) or [-1, 6)."""

  min_value: float = Field(description="The smaller end of the interval.")

  open_at_min: bool = Field(
      default=False,
      description=(
          "The interval is Open on the min end. The default value is False,"
          " which means that we assume that the interval is Closed."
      ),
  )

  max_value: float = Field(description="The larger end of the interval.")

  open_at_max: bool = Field(
      default=False,
      description=(
          "The interval is Open on the max end. The default value is False,"
          " which means that we assume that the interval is Closed."
      ),
  )


class MetricValueInfo(EvalBaseModel):
  """Information about the type of metric value."""

  interval: Optional[Interval] = Field(
      default=None,
      description="The values represented by the metric are of type interval.",
  )


class MetricInfo(EvalBaseModel):
  """Information about the metric that are used for Evals."""

  metric_name: str = Field(description="The name of the metric.")

  description: str = Field(
      default=None, description="A 2 to 3 line description of the metric."
  )

  metric_value_info: MetricValueInfo = Field(
      description="Information on the nature of values supported by the metric."
  )


class MetricInfoProvider(abc.ABC):
  """Interface for providing MetricInfo."""

  @abc.abstractmethod
  def get_metric_info(self) -> MetricInfo:
    """Returns MetricInfo for a given metric."""
    raise NotImplementedError



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Union

from google.genai import types as genai_types
from pydantic import Field
from pydantic import model_validator
from typing_extensions import TypeAlias

from .app_details import AppDetails
from .common import EvalBaseModel
from .conversation_scenarios import ConversationScenario
from .eval_rubrics import Rubric


class IntermediateData(EvalBaseModel):
  """Container for intermediate data that an agent would generate as it responds with a final answer."""

  tool_uses: list[genai_types.FunctionCall] = []
  """Tool use trajectory in chronological order."""

  tool_responses: list[genai_types.FunctionResponse] = []
  """Tool response trajectory in chronological order."""

  intermediate_responses: list[tuple[str, list[genai_types.Part]]] = []
  """Intermediate responses generated by sub-agents to convey progress or status
  in a multi-agent system, distinct from the final response.

  This is expressed as a tuple of:
    - Author: Usually the sub-agent name that generated the intermediate
      response.

    - A list of Parts that comprise of the response.
  """


class InvocationEvent(EvalBaseModel):
  """An immutable record representing a specific point in the agent's invocation.

  It captures agent's replies, requests to use tools (function calls), and tool
  results.

  This structure is a simple projection of the actual `Event` datamodel that
  is intended for the Eval System.
  """

  author: str
  """The name of the agent that authored/owned this event."""

  content: Optional[genai_types.Content]
  """The content of the event."""


class InvocationEvents(EvalBaseModel):
  """A container for events that occur during the course of an invocation."""

  invocation_events: list[InvocationEvent] = Field(default_factory=list)
  """A list of invocation events."""


IntermediateDataType: TypeAlias = Union[IntermediateData, InvocationEvents]


class Invocation(EvalBaseModel):
  """Represents a single invocation."""

  invocation_id: str = ""
  """Unique identifier for the invocation."""

  user_content: genai_types.Content
  """Content provided by the user in this invocation."""

  final_response: Optional[genai_types.Content] = None
  """Final response from the agent."""

  intermediate_data: Optional[IntermediateDataType] = None
  """Intermediate steps generated as a part of Agent execution.

  For a multi-agent system, it is also helpful to inspect the route that
  the agent took to generate final response.
  """

  creation_timestamp: float = 0.0
  """Timestamp for the current invocation, primarily intended for debugging purposes."""

  rubrics: Optional[list[Rubric]] = Field(
      default=None,
  )
  """A list of rubrics that are applicable to only this invocation."""

  app_details: Optional[AppDetails] = Field(default=None)
  """Details about the App that was used for this invocation."""


SessionState: TypeAlias = dict[str, Any]
"""The state of the session."""


class SessionInput(EvalBaseModel):
  """Values that help initialize a Session."""

  app_name: str
  """The name of the app."""

  user_id: str
  """The user id."""

  state: SessionState = Field(default_factory=dict)
  """The state of the session."""


StaticConversation: TypeAlias = list[Invocation]
"""A conversation where the user's queries for each invocation are already specified."""


class EvalCase(EvalBaseModel):
  """An eval case."""

  eval_id: str
  """Unique identifier for the evaluation case."""

  conversation: Optional[StaticConversation] = None
  """A static conversation between the user and the Agent.

   While creating an eval case you should specify either a `conversation` or a
  `conversation_scenario`, but not both.
  """

  conversation_scenario: Optional[ConversationScenario] = None
  """A conversation scenario that should be used by a UserSimulator.

  While creating an eval case you should specify either a `conversation` or a
  `conversation_scenario`, but not both.
  """

  session_input: Optional[SessionInput] = None
  """Session input that will be passed on to the Agent during eval.
     It is common for Agents state to be initialized to some initial/default value,
     for example, your agent may need to know today's date.
  """

  creation_timestamp: float = 0.0
  """The time at which this eval case was created."""

  rubrics: Optional[list[Rubric]] = Field(
      default=None,
  )
  """A list of rubrics that are applicable to all the invocations in the conversation of this eval case."""

  final_session_state: Optional[SessionState] = Field(default_factory=dict)
  """The expected final session state at the end of the conversation."""

  @model_validator(mode="after")
  def ensure_conversation_xor_conversation_scenario(self) -> EvalCase:
    if (self.conversation is None) == (self.conversation_scenario is None):
      raise ValueError(
          "Exactly one of conversation and conversation_scenario must be"
          " provided in an EvalCase."
      )
    return self


def get_all_tool_calls(
    intermediate_data: Optional[IntermediateDataType],
) -> list[genai_types.FunctionCall]:
  """A utility method to retrieve tools calls from intermediate data."""
  if not intermediate_data:
    return []

  tool_calls = []
  if isinstance(intermediate_data, IntermediateData):
    tool_calls = intermediate_data.tool_uses
  elif isinstance(intermediate_data, InvocationEvents):
    # Go over each event in the list of events
    for invocation_event in intermediate_data.invocation_events:
      # Check if the event has content and some parts.
      if invocation_event.content and invocation_event.content.parts:
        for p in invocation_event.content.parts:
          # For each part, we check if any of those part is a function call.
          if p.function_call:
            tool_calls.append(p.function_call)
  else:
    raise ValueError(
        f"Unsupported type for intermediate_data `{intermediate_data}`"
    )

  return tool_calls


def get_all_tool_responses(
    intermediate_data: Optional[IntermediateDataType],
) -> list[genai_types.FunctionResponse]:
  """A utility method to retrieve tools responses from intermediate data."""
  if not intermediate_data:
    return []

  tool_responses = []
  if isinstance(intermediate_data, IntermediateData):
    tool_responses = intermediate_data.tool_responses
  elif isinstance(intermediate_data, InvocationEvents):
    # Go over each event in the list of events
    for invocation_event in intermediate_data.invocation_events:
      # Check if the event has content and some parts.
      if invocation_event.content and invocation_event.content.parts:
        for p in invocation_event.content.parts:
          # For each part, we check if any of those part is a function response.
          if p.function_response:
            tool_responses.append(p.function_response)
  else:
    raise ValueError(
        f"Unsupported type for intermediate_data `{intermediate_data}`"
    )

  return tool_responses


ToolCallAndResponse: TypeAlias = tuple[
    genai_types.FunctionCall, Optional[genai_types.FunctionResponse]
]
"""A Tuple representing a Function call and corresponding optional function response."""


def get_all_tool_calls_with_responses(
    intermediate_data: Optional[IntermediateDataType],
) -> list[ToolCallAndResponse]:
  """Returns tool calls with the corresponding responses, if available."""
  tool_responses_by_call_id: dict[str, genai_types.FunctionResponse] = {
      tool_response.id: tool_response
      for tool_response in get_all_tool_responses(intermediate_data)
  }

  tool_call_and_responses: list[ToolCallAndResponse] = []

  for tool_call in get_all_tool_calls(intermediate_data):
    response = tool_responses_by_call_id.get(tool_call.id, None)
    tool_call_and_responses.append((tool_call, response))

  return tool_call_and_responses



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import inspect
from typing import Optional
from typing import TYPE_CHECKING
from typing import Union

from google.adk.agents.callback_context import CallbackContext
from google.adk.agents.readonly_context import ReadonlyContext
from google.adk.models.llm_request import LlmRequest
from google.adk.models.llm_response import LlmResponse
from google.adk.plugins.base_plugin import BasePlugin
from google.adk.utils import instructions_utils

if TYPE_CHECKING:
  from google.adk.agents.llm_agent import InstructionProvider
  from google.adk.agents.llm_agent import LlmAgent


class GlobalInstructionPlugin(BasePlugin):
  """Plugin that provides global instructions functionality at the App level.

  This plugin replaces the deprecated global_instruction field on LlmAgent.
  Global instructions are applied to all agents in the application, providing
  a consistent way to set application-wide instructions, identity, or
  personality.

  The plugin operates through the before_model_callback, allowing it to modify
  LLM requests before they are sent to the model.
  """

  def __init__(
      self,
      global_instruction: Union[str, InstructionProvider] = "",
      name: str = "global_instruction",
  ) -> None:
    """Initialize the GlobalInstructionPlugin.

    Args:
      global_instruction: The instruction to apply globally. Can be a string or
        an InstructionProvider function that takes ReadonlyContext and returns a
        string (sync or async).
      name: The name of the plugin (defaults to "global_instruction").
    """
    super().__init__(name=name)
    self.global_instruction = global_instruction

  async def before_model_callback(
      self, *, callback_context: CallbackContext, llm_request: LlmRequest
  ) -> Optional[LlmResponse]:
    """Apply global instructions to the LLM request.

    This callback is executed before each request is sent to the model,
    allowing the plugin to inject global instructions into the request.

    Args:
      callback_context: The context for the current agent call.
      llm_request: The prepared request object to be sent to the model.

    Returns:
      None to allow the LLM request to proceed normally.
    """
    # Only process if we have a global instruction configured
    if not self.global_instruction:
      return None

    # Resolve the global instruction (handle both string and InstructionProvider)
    final_global_instruction = await self._resolve_global_instruction(
        callback_context
    )

    if not final_global_instruction:
      return None

    # Make the global instruction the leading system instruction.
    existing_instruction = llm_request.config.system_instruction

    if not existing_instruction:
      llm_request.config.system_instruction = final_global_instruction
      return None

    if isinstance(existing_instruction, str):
      llm_request.config.system_instruction = (
          f"{final_global_instruction}\n\n{existing_instruction}"
      )
    else:  # It's an Iterable
      # Convert to list to allow prepending
      new_instruction_list = [final_global_instruction]
      new_instruction_list.extend(list(existing_instruction))
      llm_request.config.system_instruction = new_instruction_list

    return None

  async def _resolve_global_instruction(
      self, readonly_context: ReadonlyContext
  ) -> str:
    """Resolve the global instruction, handling both string and InstructionProvider.

    Args:
      readonly_context: The readonly context for resolving instructions.

    Returns:
      The fully resolved and processed global instruction string, ready to use.
    """
    if isinstance(self.global_instruction, str):
      # For string instructions, apply state injection
      return await instructions_utils.inject_session_state(
          self.global_instruction, readonly_context
      )
    else:
      # Handle InstructionProvider (callable)
      # InstructionProvider already handles state internally, no injection needed
      instruction = self.global_instruction(readonly_context)
      if inspect.isawaitable(instruction):
        instruction = await instruction
      return instruction



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import abc
from typing import Optional

from google.genai.types import Content

from ..events.event import Event
from ..utils.feature_decorator import experimental


@experimental
class BaseEventsSummarizer(abc.ABC):
  """Base interface for compacting events."""

  @abc.abstractmethod
  async def maybe_summarize_events(
      self, *, events: list[Event]
  ) -> Optional[Event]:
    """Compact a list of events into a single event.

    If compaction failed, return None. Otherwise, compact into a content and
    return it.

    This method will summarize the events and return a new summary event
    indicating the range of events it summarized.

    Args:
      events: Events to compact.

    Returns:
      The new compacted event, or None if no compaction happened.
    """
    raise NotImplementedError()



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Client library for interacting with the Google Cloud Agent Registry within ADK."""

from __future__ import annotations

from enum import Enum
import logging
import os
import re
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union
from urllib.parse import parse_qs
from urllib.parse import urlparse

from a2a.client.client_factory import minimal_agent_card
from a2a.types import AgentCapabilities
from a2a.types import AgentCard
from a2a.types import AgentSkill
from a2a.types import TransportProtocol as A2ATransport
from google.adk.agents.readonly_context import ReadonlyContext
from google.adk.agents.remote_a2a_agent import RemoteA2aAgent
from google.adk.tools.mcp_tool.mcp_session_manager import StreamableHTTPConnectionParams
from google.adk.tools.mcp_tool.mcp_toolset import McpToolset
import google.auth
import google.auth.transport.requests
import httpx

logger = logging.getLogger("google_adk." + __name__)

AGENT_REGISTRY_BASE_URL = "https://agentregistry.googleapis.com/v1alpha"


class _ProtocolType(str, Enum):
  """Supported agent protocol types."""

  TYPE_UNSPECIFIED = "TYPE_UNSPECIFIED"
  A2A_AGENT = "A2A_AGENT"
  CUSTOM = "CUSTOM"


class AgentRegistry:
  """Client for interacting with the Google Cloud Agent Registry service.

  Unlike a standard REST client library, this class provides higher-level
  abstractions for ADK integration. It surfaces the agent registry service
  methods along with helper methods like `get_mcp_toolset` and
  `get_remote_a2a_agent` that automatically resolve connection details and
  handle authentication to produce ready-to-use ADK components.
  """

  def __init__(
      self,
      project_id: Optional[str] = None,
      location: Optional[str] = None,
      header_provider: Optional[
          Callable[[ReadonlyContext], Dict[str, str]]
      ] = None,
  ):
    """Initializes the AgentRegistry client.

    Args:
      project_id: The Google Cloud project ID.
      location: The Google Cloud location (region).
      header_provider: Optional provider for custom headers.
    """
    self.project_id = project_id
    self.location = location

    if not self.project_id or not self.location:
      raise ValueError("project_id and location must be provided")

    self._base_path = f"projects/{self.project_id}/locations/{self.location}"
    self._header_provider = header_provider
    try:
      self._credentials, _ = google.auth.default()
    except google.auth.exceptions.DefaultCredentialsError as e:
      raise RuntimeError(
          f"Failed to get default Google Cloud credentials: {e}"
      ) from e

  def _get_auth_headers(self) -> Dict[str, str]:
    """Refreshes credentials and returns authorization headers."""
    try:
      request = google.auth.transport.requests.Request()
      self._credentials.refresh(request)
      headers = {
          "Authorization": f"Bearer {self._credentials.token}",
          "Content-Type": "application/json",
      }
      quota_project_id = getattr(self._credentials, "quota_project_id", None)
      if quota_project_id:
        headers["x-goog-user-project"] = quota_project_id
      return headers
    except google.auth.exceptions.RefreshError as e:
      raise RuntimeError(
          f"Failed to refresh Google Cloud credentials: {e}"
      ) from e

  def _make_request(
      self, path: str, params: Optional[Dict[str, Any]] = None
  ) -> Dict[str, Any]:
    """Helper function to make GET requests to the Agent Registry API."""
    if path.startswith("projects/"):
      url = f"{AGENT_REGISTRY_BASE_URL}/{path}"
    else:
      url = f"{AGENT_REGISTRY_BASE_URL}/{self._base_path}/{path}"

    try:
      headers = self._get_auth_headers()
      with httpx.Client() as client:
        response = client.get(url, headers=headers, params=params)
        response.raise_for_status()
        return response.json()
    except httpx.HTTPStatusError as e:
      raise RuntimeError(
          f"API request failed with status {e.response.status_code}:"
          f" {e.response.text}"
      ) from e
    except httpx.RequestError as e:
      raise RuntimeError(f"API request failed (network error): {e}") from e
    except Exception as e:
      raise RuntimeError(f"API request failed: {e}") from e

  def _get_connection_uri(
      self,
      resource_details: Dict[str, Any],
      protocol_type: Optional[_ProtocolType] = None,
      protocol_binding: Optional[A2ATransport] = None,
  ) -> Optional[str]:
    """Extracts the first matching URI based on type and binding filters."""
    protocols = list(resource_details.get("protocols", []))
    if "interfaces" in resource_details:
      protocols.append({"interfaces": resource_details["interfaces"]})

    for p in protocols:
      if protocol_type and p.get("type") != protocol_type:
        continue
      for i in p.get("interfaces", []):
        if protocol_binding and i.get("protocolBinding") != protocol_binding:
          continue
        if url := i.get("url"):
          return url

    return None

  def _clean_name(self, name: str) -> str:
    """Cleans a string to be a valid Python identifier for agent names."""
    clean = re.sub(r"[^a-zA-Z0-9_]", "_", name)
    clean = re.sub(r"_+", "_", clean)
    clean = clean.strip("_")
    if clean and not clean[0].isalpha() and clean[0] != "_":
      clean = "_" + clean
    return clean

  # --- MCP Server Methods ---

  def list_mcp_servers(
      self,
      filter_str: Optional[str] = None,
      page_size: Optional[int] = None,
      page_token: Optional[str] = None,
  ) -> Dict[str, Any]:
    """Fetches a list of MCP Servers."""
    params = {}
    if filter_str:
      params["filter"] = filter_str
    if page_size:
      params["pageSize"] = str(page_size)
    if page_token:
      params["pageToken"] = page_token
    return self._make_request("mcpServers", params=params)

  def get_mcp_server(self, name: str) -> Dict[str, Any]:
    """Retrieves details of a specific MCP Server."""
    return self._make_request(name)

  def get_mcp_toolset(self, mcp_server_name: str) -> McpToolset:
    """Constructs an McpToolset instance from a registered MCP Server."""
    server_details = self.get_mcp_server(mcp_server_name)
    name = self._clean_name(server_details.get("displayName", mcp_server_name))

    endpoint_uri = self._get_connection_uri(
        server_details, protocol_binding=A2ATransport.jsonrpc
    ) or self._get_connection_uri(
        server_details, protocol_binding=A2ATransport.http_json
    )
    if not endpoint_uri:
      raise ValueError(
          f"MCP Server endpoint URI not found for: {mcp_server_name}"
      )

    connection_params = StreamableHTTPConnectionParams(
        url=endpoint_uri, headers=self._get_auth_headers()
    )
    return McpToolset(
        connection_params=connection_params,
        tool_name_prefix=name,
        header_provider=self._header_provider,
    )

  # --- Agent Methods ---

  def list_agents(
      self,
      filter_str: Optional[str] = None,
      page_size: Optional[int] = None,
      page_token: Optional[str] = None,
  ) -> Dict[str, Any]:
    """Fetches a list of registered A2A Agents."""
    params = {}
    if filter_str:
      params["filter"] = filter_str
    if page_size:
      params["pageSize"] = str(page_size)
    if page_token:
      params["pageToken"] = page_token
    return self._make_request("agents", params=params)

  def get_agent_info(self, name: str) -> Dict[str, Any]:
    """Retrieves detailed metadata of a specific A2A Agent."""
    return self._make_request(name)

  def get_remote_a2a_agent(self, agent_name: str) -> RemoteA2aAgent:
    """Creates a RemoteA2aAgent instance for a registered A2A Agent."""
    agent_info = self.get_agent_info(agent_name)

    # Try to use the full agent card if available
    card = agent_info.get("card", {})
    card_content = card.get("content")
    if card.get("type") == "A2A_AGENT_CARD" and card_content:
      agent_card = AgentCard(**card_content)
      # Clean the name to be a valid identifier
      name = self._clean_name(agent_card.name)
      return RemoteA2aAgent(
          name=name,
          agent_card=agent_card,
          description=agent_card.description,
      )

    name = self._clean_name(agent_info.get("displayName", agent_name))
    description = agent_info.get("description", "")
    version = agent_info.get("version", "")

    url = self._get_connection_uri(
        agent_info, protocol_type=_ProtocolType.A2A_AGENT
    )
    if not url:
      raise ValueError(f"A2A connection URI not found for Agent: {agent_name}")

    skills = []
    for s in agent_info.get("skills", []):
      skills.append(
          AgentSkill(
              id=s.get("id"),
              name=s.get("name"),
              description=s.get("description", ""),
              tags=s.get("tags", []),
              examples=s.get("examples", []),
          )
      )

    agent_card = AgentCard(
        name=name,
        description=description,
        version=version,
        url=url,
        skills=skills,
        capabilities=AgentCapabilities(streaming=False, polling=False),
        defaultInputModes=["text"],
        defaultOutputModes=["text"],
    )

    return RemoteA2aAgent(
        name=name,
        agent_card=agent_card,
        description=description,
    )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from contextlib import asynccontextmanager
import copy
import json
import logging
import os
import sqlite3
from typing import Any
from typing import Optional
from urllib.parse import unquote
from urllib.parse import urlparse

import aiosqlite
from google.adk.platform import time as platform_time
from google.adk.platform import uuid as platform_uuid
from typing_extensions import override

from . import _session_util
from ..errors.already_exists_error import AlreadyExistsError
from ..events.event import Event
from .base_session_service import BaseSessionService
from .base_session_service import GetSessionConfig
from .base_session_service import ListSessionsResponse
from .session import Session
from .state import State

logger = logging.getLogger("google_adk." + __name__)

PRAGMA_FOREIGN_KEYS = "PRAGMA foreign_keys = ON"

APP_STATES_TABLE_SCHEMA = """
CREATE TABLE IF NOT EXISTS app_states (
    app_name TEXT PRIMARY KEY,
    state TEXT NOT NULL,
    update_time REAL NOT NULL
);
"""

USER_STATES_TABLE_SCHEMA = """
CREATE TABLE IF NOT EXISTS user_states (
    app_name TEXT NOT NULL,
    user_id TEXT NOT NULL,
    state TEXT NOT NULL,
    update_time REAL NOT NULL,
    PRIMARY KEY (app_name, user_id)
);
"""

SESSIONS_TABLE_SCHEMA = """
CREATE TABLE IF NOT EXISTS sessions (
    app_name TEXT NOT NULL,
    user_id TEXT NOT NULL,
    id TEXT NOT NULL,
    state TEXT NOT NULL,
    create_time REAL NOT NULL,
    update_time REAL NOT NULL,
    PRIMARY KEY (app_name, user_id, id)
);
"""

EVENTS_TABLE_SCHEMA = """
CREATE TABLE IF NOT EXISTS events (
    id TEXT NOT NULL,
    app_name TEXT NOT NULL,
    user_id TEXT NOT NULL,
    session_id TEXT NOT NULL,
    invocation_id TEXT NOT NULL,
    timestamp REAL NOT NULL,
    event_data TEXT NOT NULL,
    PRIMARY KEY (app_name, user_id, session_id, id),
    FOREIGN KEY (app_name, user_id, session_id) REFERENCES sessions(app_name, user_id, id) ON DELETE CASCADE
);
"""
CREATE_SCHEMA_SQL = "\n".join([
    APP_STATES_TABLE_SCHEMA,
    USER_STATES_TABLE_SCHEMA,
    SESSIONS_TABLE_SCHEMA,
    EVENTS_TABLE_SCHEMA,
])


def _parse_db_path(db_path: str) -> tuple[str, str, bool]:
  """Normalizes a SQLite db path from a URL or filesystem path.

  Returns:
    A tuple of:
      - filesystem path (for `os.path.exists` and user-facing messages)
      - value to pass to sqlite/aiosqlite connect
      - whether to pass `uri=True` to sqlite/aiosqlite connect

  Notes:
    When a SQLAlchemy-style SQLite URL is provided, this follows SQLAlchemy's
    conventions:
      - `sqlite:///relative.db` is a path relative to the current working dir.
      - `sqlite:////absolute.db` is an absolute filesystem path.
  """
  if not db_path.startswith(("sqlite:", "sqlite+aiosqlite:")):
    return db_path, db_path, False

  parsed = urlparse(db_path)
  raw_path = unquote(parsed.path)
  if not raw_path:
    return db_path, db_path, False

  normalized_path = raw_path
  if normalized_path.startswith("//"):
    normalized_path = normalized_path[1:]
  elif normalized_path.startswith("/"):
    normalized_path = normalized_path[1:]

  if parsed.query:
    # sqlite3 only treats the filename as a URI when it starts with `file:`.
    return normalized_path, f"file:{normalized_path}?{parsed.query}", True

  return normalized_path, normalized_path, False


class SqliteSessionService(BaseSessionService):
  """A session service that uses an SQLite database for storage via aiosqlite.

  Event data is stored as JSON to allow for schema flexibility as event
  fields evolve.
  """

  def __init__(self, db_path: str):
    """Initializes the SQLite session service with a database path."""
    self._db_path, self._db_connect_path, self._db_connect_uri = _parse_db_path(
        db_path
    )

    if self._is_migration_needed():
      raise RuntimeError(
          f"Database {self._db_path} seems to use an old schema."
          " Please run the migration command to"
          " migrate it to the new schema. Example: `python -m"
          " google.adk.sessions.migration.migrate_from_sqlalchemy_sqlite"
          f" --source_db_path {self._db_path} --dest_db_path"
          f" {self._db_path}.new` then backup {self._db_path} and rename"
          f" {self._db_path}.new to {self._db_path}."
      )

  @override
  async def create_session(
      self,
      *,
      app_name: str,
      user_id: str,
      state: Optional[dict[str, Any]] = None,
      session_id: Optional[str] = None,
  ) -> Session:
    if session_id:
      session_id = session_id.strip()
    if not session_id:
      session_id = platform_uuid.new_uuid()
    now = platform_time.get_time()

    async with self._get_db_connection() as db:
      # Check if session_id already exists
      async with db.execute(
          "SELECT 1 FROM sessions WHERE app_name=? AND user_id=? AND id=?",
          (app_name, user_id, session_id),
      ) as cursor:
        if await cursor.fetchone():
          raise AlreadyExistsError(
              f"Session with id {session_id} already exists."
          )

      # Extract state deltas
      state_deltas = _session_util.extract_state_delta(state)
      app_state_delta = state_deltas["app"]
      user_state_delta = state_deltas["user"]
      session_state = state_deltas["session"]

      # Apply state delta and update/insert states atomically
      if app_state_delta:
        await self._upsert_app_state(db, app_name, app_state_delta, now)
      if user_state_delta:
        await self._upsert_user_state(
            db, app_name, user_id, user_state_delta, now
        )

      # Fetch current state after upserts
      storage_app_state = await self._get_app_state(db, app_name)
      storage_user_state = await self._get_user_state(db, app_name, user_id)

      # Store the session
      await db.execute(
          """
          INSERT INTO sessions (app_name, user_id, id, state, create_time, update_time)
          VALUES (?, ?, ?, ?, ?, ?)
          """,
          (
              app_name,
              user_id,
              session_id,
              json.dumps(session_state),
              now,
              now,
          ),
      )
      await db.commit()

      # Merge states for response
      merged_state = _merge_state(
          storage_app_state, storage_user_state, session_state
      )
      return Session(
          app_name=app_name,
          user_id=user_id,
          id=session_id,
          state=merged_state,
          events=[],
          last_update_time=now,
      )

  @override
  async def get_session(
      self,
      *,
      app_name: str,
      user_id: str,
      session_id: str,
      config: Optional[GetSessionConfig] = None,
  ) -> Optional[Session]:
    async with self._get_db_connection() as db:
      async with db.execute(
          "SELECT state, update_time FROM sessions WHERE app_name=? AND"
          " user_id=? AND id=?",
          (app_name, user_id, session_id),
      ) as cursor:
        session_row = await cursor.fetchone()
        if session_row is None:
          return None
        session_state = json.loads(session_row["state"])
        last_update_time = session_row["update_time"]

      # Build events query
      query_parts = [
          "SELECT event_data FROM events",
          "WHERE app_name=? AND user_id=? AND session_id=?",
      ]
      params: list[Any] = [app_name, user_id, session_id]

      if config and config.after_timestamp:
        query_parts.append("AND timestamp >= ?")
        params.append(config.after_timestamp)

      query_parts.append("ORDER BY timestamp DESC")

      if config and config.num_recent_events:
        query_parts.append("LIMIT ?")
        params.append(config.num_recent_events)

      event_rows = await db.execute_fetchall(" ".join(query_parts), params)
      storage_events_data = [row["event_data"] for row in event_rows]

      # Fetch states from storage
      app_state = await self._get_app_state(db, app_name)
      user_state = await self._get_user_state(db, app_name, user_id)

      # Merge states
      merged_state = _merge_state(app_state, user_state, session_state)

      # Deserialize events and reverse to chronological order
      events = [
          Event.model_validate_json(event_data)
          for event_data in reversed(storage_events_data)
      ]

      return Session(
          app_name=app_name,
          user_id=user_id,
          id=session_id,
          state=merged_state,
          events=events,
          last_update_time=last_update_time,
      )

  @override
  async def list_sessions(
      self, *, app_name: str, user_id: Optional[str] = None
  ) -> ListSessionsResponse:
    sessions_list = []
    async with self._get_db_connection() as db:
      # Fetch sessions
      if user_id:
        session_rows = await db.execute_fetchall(
            "SELECT id, user_id, state, update_time FROM sessions WHERE"
            " app_name=? AND user_id=?",
            (app_name, user_id),
        )
      else:
        session_rows = await db.execute_fetchall(
            "SELECT id, user_id, state, update_time FROM sessions WHERE"
            " app_name=?",
            (app_name,),
        )

      # Fetch app state
      app_state = await self._get_app_state(db, app_name)

      # Fetch user states
      user_states_map = {}
      if user_id:
        user_state = await self._get_user_state(db, app_name, user_id)
        if user_state:
          user_states_map[user_id] = user_state
      else:
        async with db.execute(
            "SELECT user_id, state FROM user_states WHERE app_name=?",
            (app_name,),
        ) as cursor:
          async for row in cursor:
            user_states_map[row["user_id"]] = json.loads(row["state"])

      # Build session list
      for row in session_rows:
        session_user_id = row["user_id"]
        session_state = json.loads(row["state"])
        user_state = user_states_map.get(session_user_id, {})
        merged_state = _merge_state(app_state, user_state, session_state)
        sessions_list.append(
            Session(
                app_name=app_name,
                user_id=session_user_id,
                id=row["id"],
                state=merged_state,
                events=[],
                last_update_time=row["update_time"],
            )
        )
    return ListSessionsResponse(sessions=sessions_list)

  @override
  async def delete_session(
      self, *, app_name: str, user_id: str, session_id: str
  ) -> None:
    async with self._get_db_connection() as db:
      await db.execute(
          "DELETE FROM sessions WHERE app_name=? AND user_id=? AND id=?",
          (app_name, user_id, session_id),
      )
      await db.commit()

  @override
  async def append_event(self, session: Session, event: Event) -> Event:
    if event.partial:
      return event

    # Apply temp state to in-memory session before trimming, so that
    # subsequent agents within the same invocation can read temp values.
    self._apply_temp_state(session, event)
    # Trim temp state before persisting
    event = self._trim_temp_delta_state(event)
    event_timestamp = event.timestamp

    async with self._get_db_connection() as db:
      # Check for stale session
      async with db.execute(
          "SELECT update_time FROM sessions WHERE app_name=? AND user_id=? AND"
          " id=?",
          (session.app_name, session.user_id, session.id),
      ) as cursor:
        row = await cursor.fetchone()
        if row is None:
          raise ValueError(f"Session {session.id} not found.")
        storage_update_time = row["update_time"]
        if storage_update_time > session.last_update_time:
          raise ValueError(
              "The last_update_time provided in the session object is"
              " earlier than the update_time in storage."
              " Please check if it is a stale session."
          )

      # Apply state delta if present
      has_session_state_delta = False
      if event.actions and event.actions.state_delta:
        state_deltas = _session_util.extract_state_delta(
            event.actions.state_delta
        )
        app_state_delta = state_deltas["app"]
        user_state_delta = state_deltas["user"]
        session_state_delta = state_deltas["session"]

        if app_state_delta:
          await self._upsert_app_state(
              db, session.app_name, app_state_delta, event_timestamp
          )
        if user_state_delta:
          await self._upsert_user_state(
              db,
              session.app_name,
              session.user_id,
              user_state_delta,
              event_timestamp,
          )
        if session_state_delta:
          await self._update_session_state_in_db(
              db,
              session.app_name,
              session.user_id,
              session.id,
              session_state_delta,
              event_timestamp,
          )
          has_session_state_delta = True

      # Insert event and update session timestamp
      await db.execute(
          """
          INSERT INTO events (id, app_name, user_id, session_id, invocation_id, timestamp, event_data)
          VALUES (?, ?, ?, ?, ?, ?, ?)
          """,
          (
              event.id,
              session.app_name,
              session.user_id,
              session.id,
              event.invocation_id,
              event.timestamp,
              event.model_dump_json(exclude_none=True),
          ),
      )
      if not has_session_state_delta:
        await db.execute(
            "UPDATE sessions SET update_time=? WHERE app_name=? AND user_id=?"
            " AND id=?",
            (
                event_timestamp,
                session.app_name,
                session.user_id,
                session.id,
            ),
        )
      await db.commit()

      # Update timestamp based on event time
      session.last_update_time = event_timestamp

    # Also update the in-memory session
    await super().append_event(session=session, event=event)
    return event

  @asynccontextmanager
  async def _get_db_connection(self):
    """Connects to the db and performs initial setup."""
    async with aiosqlite.connect(
        self._db_connect_path, uri=self._db_connect_uri
    ) as db:
      db.row_factory = aiosqlite.Row
      await db.execute(PRAGMA_FOREIGN_KEYS)
      await db.executescript(CREATE_SCHEMA_SQL)
      yield db

  async def _get_state(
      self, db: aiosqlite.Connection, query: str, params: tuple
  ) -> dict[str, Any]:
    """Fetches and deserializes a JSON state column from a single row."""
    async with db.execute(query, params) as cursor:
      row = await cursor.fetchone()
      return json.loads(row["state"]) if row else {}

  async def _get_app_state(
      self, db: aiosqlite.Connection, app_name: str
  ) -> dict[str, Any]:
    return await self._get_state(
        db, "SELECT state FROM app_states WHERE app_name=?", (app_name,)
    )

  async def _get_user_state(
      self, db: aiosqlite.Connection, app_name: str, user_id: str
  ) -> dict[str, Any]:
    return await self._get_state(
        db,
        "SELECT state FROM user_states WHERE app_name=? AND user_id=?",
        (app_name, user_id),
    )

  async def _get_session_state(
      self,
      db: aiosqlite.Connection,
      app_name: str,
      user_id: str,
      session_id: str,
  ) -> dict[str, Any]:
    return await self._get_state(
        db,
        "SELECT state FROM sessions WHERE app_name=? AND user_id=? AND id=?",
        (app_name, user_id, session_id),
    )

  async def _upsert_app_state(
      self, db: aiosqlite.Connection, app_name: str, delta: dict, now: float
  ) -> None:
    """Atomically inserts or updates app state using json_patch."""
    await db.execute(
        """
        INSERT INTO app_states (app_name, state, update_time) VALUES (?, ?, ?)
        ON CONFLICT(app_name) DO UPDATE SET state=json_patch(state, excluded.state), update_time=excluded.update_time
        """,
        (app_name, json.dumps(delta), now),
    )

  async def _upsert_user_state(
      self,
      db: aiosqlite.Connection,
      app_name: str,
      user_id: str,
      delta: dict,
      now: float,
  ) -> None:
    """Atomically inserts or updates user state using json_patch."""
    await db.execute(
        """
        INSERT INTO user_states (app_name, user_id, state, update_time) VALUES (?, ?, ?, ?)
        ON CONFLICT(app_name, user_id) DO UPDATE SET state=json_patch(state, excluded.state), update_time=excluded.update_time
        """,
        (app_name, user_id, json.dumps(delta), now),
    )

  async def _update_session_state_in_db(
      self,
      db: aiosqlite.Connection,
      app_name: str,
      user_id: str,
      session_id: str,
      delta: dict,
      now: float,
  ) -> None:
    """Atomically updates session state using json_patch."""
    await db.execute(
        "UPDATE sessions SET state=json_patch(state, ?), update_time=? WHERE"
        " app_name=? AND user_id=? AND id=?",
        (
            json.dumps(delta),
            now,
            app_name,
            user_id,
            session_id,
        ),
    )

  def _is_migration_needed(self) -> bool:
    """Checks if migration to new schema is needed."""
    if not os.path.exists(self._db_path):
      return False
    try:
      with sqlite3.connect(
          self._db_connect_path, uri=self._db_connect_uri
      ) as conn:
        cursor = conn.cursor()
        # Check if events table exists
        cursor.execute(
            "SELECT 1 FROM sqlite_master WHERE type='table' and name='events'"
        )
        if not cursor.fetchone():
          return False  # No events table, so no migration needed.

        # If events table exists, check for event_data column
        cursor.execute("PRAGMA table_info(events)")
        columns = [row[1] for row in cursor.fetchall()]
        if "event_data" in columns:
          return False  # New schema: event_data column exists.
        else:
          return (
              True  # Old schema: events table exists, but no event_data column.
          )
    except sqlite3.Error as e:
      raise RuntimeError(
          f"Error accessing database {self._db_path}: {e}"
      ) from e


def _merge_state(app_state, user_state, session_state):
  """Merges app, user, and session states into a single dictionary."""
  merged_state = copy.deepcopy(session_state)
  for key, value in app_state.items():
    merged_state[State.APP_PREFIX + key] = value
  for key, value in user_state.items():
    merged_state[State.USER_PREFIX + key] = value
  return merged_state



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for Agent Skills."""

from __future__ import annotations

import logging
import pathlib
from typing import Union

from google.auth import credentials as auth
from google.cloud import storage
from pydantic import ValidationError
import yaml

from . import models

_ALLOWED_FRONTMATTER_KEYS = frozenset({
    "name",
    "description",
    "license",
    "allowed-tools",
    "allowed_tools",
    "metadata",
    "compatibility",
})


def _load_dir(directory: pathlib.Path) -> dict[str, str]:
  """Recursively load files from a directory into a dictionary.

  Args:
    directory: Path to the directory to load.

  Returns:
    Dictionary mapping relative file paths to their string content.
  """
  files = {}
  if directory.exists() and directory.is_dir():
    for file_path in directory.rglob("*"):
      if "__pycache__" in file_path.parts:
        continue
      if file_path.is_file():
        relative_path = file_path.relative_to(directory)
        try:
          files[str(relative_path)] = file_path.read_text(encoding="utf-8")
        except UnicodeDecodeError:
          # Binary files or non-UTF-8 files are skipped for text content.
          continue
  return files


def _parse_skill_md_content(content: str) -> tuple[dict, str]:
  """Parse SKILL.md from raw content string.

  Args:
    content: The string content of SKILL.md.

  Returns:
    Tuple of (parsed_frontmatter_dict, body_string).

  Raises:
    ValueError: If SKILL.md is invalid.
  """
  if not content.startswith("---"):
    raise ValueError("SKILL.md must start with YAML frontmatter (---)")

  parts = content.split("---", 2)
  if len(parts) < 3:
    raise ValueError("SKILL.md frontmatter not properly closed with ---")

  frontmatter_str = parts[1]
  body = parts[2].strip()

  try:
    parsed = yaml.safe_load(frontmatter_str)
  except yaml.YAMLError as e:
    raise ValueError(f"Invalid YAML in frontmatter: {e}") from e

  if not isinstance(parsed, dict):
    raise ValueError("SKILL.md frontmatter must be a YAML mapping")

  return parsed, body


def _parse_skill_md(
    skill_dir: pathlib.Path,
) -> tuple[dict, str, pathlib.Path]:
  """Parse SKILL.md from a skill directory.

  Args:
    skill_dir: Resolved path to the skill directory.

  Returns:
    Tuple of (parsed_frontmatter_dict, body_string, skill_md_path).

  Raises:
    FileNotFoundError: If the directory or SKILL.md is not found.
    ValueError: If SKILL.md is invalid.
  """
  if not skill_dir.is_dir():
    raise FileNotFoundError(f"Skill directory '{skill_dir}' not found.")

  skill_md = None
  for name in ("SKILL.md", "skill.md"):
    path = skill_dir / name
    if path.exists():
      skill_md = path
      break

  if skill_md is None:
    raise FileNotFoundError(f"SKILL.md not found in '{skill_dir}'.")

  content = skill_md.read_text(encoding="utf-8")
  parsed, body = _parse_skill_md_content(content)

  return parsed, body, skill_md


def _load_skill_from_dir(skill_dir: Union[str, pathlib.Path]) -> models.Skill:
  """Load a complete skill from a directory.

  Args:
    skill_dir: Path to the skill directory.

  Returns:
    Skill object with all components loaded.

  Raises:
    FileNotFoundError: If the skill directory or SKILL.md is not found.
    ValueError: If SKILL.md is invalid or the skill name does not match
      the directory name.
  """
  skill_dir = pathlib.Path(skill_dir).resolve()

  parsed, body, skill_md = _parse_skill_md(skill_dir)

  # Use model_validate to handle aliases like allowed-tools
  frontmatter = models.Frontmatter.model_validate(parsed)

  # Validate that skill name matches the directory name
  if skill_dir.name != frontmatter.name:
    raise ValueError(
        f"Skill name '{frontmatter.name}' does not match directory"
        f" name '{skill_dir.name}'."
    )

  references = _load_dir(skill_dir / "references")
  assets = _load_dir(skill_dir / "assets")
  raw_scripts = _load_dir(skill_dir / "scripts")
  scripts = {
      name: models.Script(src=content) for name, content in raw_scripts.items()
  }

  resources = models.Resources(
      references=references,
      assets=assets,
      scripts=scripts,
  )

  return models.Skill(
      frontmatter=frontmatter,
      instructions=body,
      resources=resources,
  )


def _validate_skill_dir(
    skill_dir: Union[str, pathlib.Path],
) -> list[str]:
  """Validate a skill directory without fully loading it.

  Checks that the directory exists, contains a valid SKILL.md with correct
  frontmatter, and that the skill name matches the directory name.

  Args:
    skill_dir: Path to the skill directory.

  Returns:
    List of problem strings. Empty list means the skill is valid.
  """
  problems: list[str] = []
  skill_dir = pathlib.Path(skill_dir).resolve()

  if not skill_dir.exists():
    return [f"Directory '{skill_dir}' does not exist."]
  if not skill_dir.is_dir():
    return [f"'{skill_dir}' is not a directory."]

  skill_md = None
  for name in ("SKILL.md", "skill.md"):
    path = skill_dir / name
    if path.exists():
      skill_md = path
      break
  if skill_md is None:
    return [f"SKILL.md not found in '{skill_dir}'."]

  try:
    parsed, _, _ = _parse_skill_md(skill_dir)
  except (FileNotFoundError, ValueError) as e:
    return [str(e)]

  unknown = set(parsed.keys()) - _ALLOWED_FRONTMATTER_KEYS
  if unknown:
    problems.append(f"Unknown frontmatter fields: {sorted(unknown)}")

  try:
    frontmatter = models.Frontmatter.model_validate(parsed)
  except ValidationError as e:
    problems.append(f"Frontmatter validation error: {e}")
    return problems

  if skill_dir.name != frontmatter.name:
    problems.append(
        f"Skill name '{frontmatter.name}' does not match directory"
        f" name '{skill_dir.name}'."
    )

  return problems


def _read_skill_properties(
    skill_dir: Union[str, pathlib.Path],
) -> models.Frontmatter:
  """Read only the frontmatter properties from a skill directory.

  This is a lightweight alternative to ``load_skill_from_dir`` when you
  only need the skill metadata without loading instructions or resources.

  Args:
    skill_dir: Path to the skill directory.

  Returns:
    Frontmatter object with the skill's metadata.

  Raises:
    FileNotFoundError: If the directory or SKILL.md is not found.
    ValueError: If the frontmatter is invalid.
  """
  skill_dir = pathlib.Path(skill_dir).resolve()
  parsed, _, _ = _parse_skill_md(skill_dir)
  return models.Frontmatter.model_validate(parsed)


def _list_skills_in_dir(
    skills_base_path: Union[str, pathlib.Path],
) -> dict[str, models.Frontmatter]:
  """List skills in a local directory.

  Args:
    skills_base_path: Path to the base directory containing skills.

  Returns:
    Dictionary mapping skill IDs to their frontmatter.
  """
  skills_base_path = pathlib.Path(skills_base_path).resolve()
  skills = {}

  if not skills_base_path.is_dir():
    logging.warning(
        "Skills base path '%s' is not a directory.", skills_base_path
    )
    return skills

  for skill_dir in sorted(skills_base_path.iterdir()):
    if not skill_dir.is_dir():
      continue

    skill_id = skill_dir.name
    try:
      frontmatter = _read_skill_properties(skill_dir)
      if skill_id != frontmatter.name:
        raise ValueError(
            f"Skill name '{frontmatter.name}' does not match directory"
            f" name '{skill_id}'."
        )
      skills[skill_id] = frontmatter
    except (FileNotFoundError, ValueError, ValidationError) as e:
      # log invalid skills during listing and skip them
      logging.warning(
          "Skipping invalid skill '%s' in directory '%s': %s",
          skill_id,
          skills_base_path,
          e,
      )
  return skills


def _list_skills_in_gcs_dir(
    bucket_name: str,
    skills_base_path: str = "",
    project_id: str | None = None,
    credentials: auth.Credentials | None = None,
) -> Dict[str, models.Frontmatter]:
  """List skills in a GCS directory.

  Args:
    bucket_name: Name of the GCS bucket.
    skills_base_path: Base directory within the bucket (e.g., 'path/to/skills').

  Returns:
    Dictionary mapping skill IDs to their frontmatter.
  """
  client = storage.Client(project=project_id, credentials=credentials)
  bucket = client.bucket(bucket_name)

  base_prefix = skills_base_path.strip("/")
  if base_prefix:
    base_prefix += "/"

  iterator = bucket.list_blobs(prefix=base_prefix, delimiter="/")
  # We must consume the iterator to populate the prefixes attribute
  for _ in iterator:
    pass
  logging.info("Found %s skills in GCS.", iterator.prefixes)

  skills = {}
  for skill_prefix in sorted(iterator.prefixes):
    manifest_blob = bucket.blob(f"{skill_prefix}SKILL.md")

    if manifest_blob.exists():
      content = manifest_blob.download_as_text()
      skill_id = skill_prefix.rstrip("/").split("/")[-1]
      try:
        parsed, _ = _parse_skill_md_content(content)
        frontmatter = models.Frontmatter.model_validate(parsed)
        skills[skill_id] = frontmatter
      except (ValueError, ValidationError) as e:
        # log invalid skills during listing and skip them
        logging.warning(
            "Skipping invalid skill '%s' in bucket '%s': %s",
            skill_id,
            bucket_name,
            e,
        )
  return skills


def _load_skill_from_gcs_dir(
    bucket_name: str,
    skill_id: str,
    skills_base_path: str = "",
    project_id: str | None = None,
    credentials: auth.Credentials | None = None,
) -> models.Skill:
  """Load a complete skill from a GCS directory.

  Args:
    bucket_name: Name of the GCS bucket.
    skill_id: The ID of the skill (directory name).
    skills_base_path: Base directory within the bucket (e.g., 'path/to/skills').
    project_id: Project ID to use for GCS client.
    credentials: Credentials to use for GCS client.

  Returns:
    Skill object with all components loaded.

  Raises:
    FileNotFoundError: If the skill directory or SKILL.md is not found.
    ValueError: If SKILL.md is invalid or the skill name does not match
      the directory name.
  """

  client = storage.Client(project=project_id, credentials=credentials)
  bucket = client.bucket(bucket_name)

  base_prefix = skills_base_path.strip("/")
  if base_prefix:
    base_prefix += "/"

  skill_dir_prefix = f"{base_prefix}{skill_id}/"
  manifest_blob = bucket.blob(f"{skill_dir_prefix}SKILL.md")

  if not manifest_blob.exists():
    raise FileNotFoundError(
        f"SKILL.md not found at gs://{bucket_name}/{skill_dir_prefix}SKILL.md"
    )

  content = manifest_blob.download_as_text()
  parsed, body = _parse_skill_md_content(content)
  frontmatter = models.Frontmatter.model_validate(parsed)

  # Validate that skill name matches the directory name
  skill_name_expected = skill_id.strip("/").split("/")[-1]
  if skill_name_expected != frontmatter.name:
    raise ValueError(
        f"Skill name '{frontmatter.name}' does not match directory"
        f" name '{skill_name_expected}'."
    )

  def _load_files_in_dir(subdir: str) -> Dict[str, Union[str, bytes]]:
    prefix = f"{skill_dir_prefix}{subdir}/"
    blobs = bucket.list_blobs(prefix=prefix)
    result = {}

    for blob in blobs:
      relative_path = blob.name[len(prefix) :]
      if not relative_path:
        continue

      try:
        result[relative_path] = blob.download_as_text()
      except UnicodeDecodeError:
        result[relative_path] = blob.download_as_bytes()
    return result

  references = _load_files_in_dir("references")
  assets = _load_files_in_dir("assets")
  raw_scripts = _load_files_in_dir("scripts")

  scripts = {}
  for name, src in raw_scripts.items():
    if isinstance(src, bytes):
      try:
        src = src.decode("utf-8")
      except UnicodeDecodeError:
        continue  # skip binary scripts if any
    scripts[name] = models.Script(src=src)

  resources = models.Resources(
      references=references,
      assets=assets,
      scripts=scripts,
  )

  return models.Skill(
      frontmatter=frontmatter,
      instructions=body,
      resources=resources,
  )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import textwrap
from typing import Optional

from .user_simulator_personas import UserPersona

_DEFAULT_USER_SIMULATOR_INSTRUCTIONS_TEMPLATE = """You are a Simulated User designed to test an AI Agent.

Your single most important job is to react logically to the Agent's last message.
The Conversation Plan is your canonical grounding, not a script; your response MUST be dictated by what the Agent just said.

# Primary Operating Loop

You MUST follow this three-step process while thinking:

Step 1: Analyze what the Agent just said or did. Specifically, is the Agent asking you a question, reporting a successful or unsuccessful operation, or saying something incorrect or unexpected?

Step 2: Choose one action based on your analysis:
* ANSWER any questions the Agent asked.
* ADVANCE to the next request as per the Conversation Plan if the Agent succeeds in satisfying your current request.
* INTERVENE if the Agent is yet to complete your current request and the Conversation Plan requires you to modify it.
* CORRECT the Agent if it is making a mistake or failing.
* END the conversation if any of the below stopping conditions are met:
  - The Agent has completed all your requests from the Conversation Plan.
  - The Agent has failed to fulfill a request *more than once*.
  - The Agent has performed an incorrect operation and informs you that it is unable to correct it.
  - The Agent ends the conversation on its own by transferring you to a *human/live agent* (NOT another AI Agent).

Step 3: Formulate a response based on the chosen action and the below Action Protocols and output it.

# Action Protocols

**PROTOCOL: ANSWER**
* Only answer the Agent's questions using information from the Conversation Plan.
* Do NOT provide any additional information the Agent did not explicitly ask for.
* If you do not have the information requested by the Agent, inform the Agent. Do NOT make up information that is not in the Conversation Plan.
* Do NOT advance to the next request in the Conversation Plan.

**PROTOCOL: ADVANCE**
* Make the next request from the Conversation Plan.
* Skip redundant requests already fulfilled by the Agent.

**PROTOCOL: INTERVENE**
* Change your current request as directed by the Conversation Plan with natural phrasing.

**PROTOCOL: CORRECT**
* Challenge illogical or incorrect statements made by the Agent.
* If the Agent did an incorrect operation, ask the Agent to fix it.
* If this is the FIRST time the Agent failed to satisfy your request, ask the Agent to try again.

**PROTOCOL: END**
* End the conversation only when any of the stopping conditions are met; do NOT end prematurely.
* Output `{{ stop_signal }}` to indicate that the conversation with the AI Agents is over.

# Conversation Plan

{{ conversation_plan }}

# Conversation History

{{ conversation_history }}
"""

_USER_SIMULATOR_INSTRUCTIONS_WITH_PERSONA_TEMPLATE = """
You are a Simulated User designed to test an AI Agent.

Your single most important job is to react logically to the Agent's last message while role-playing as the given Persona.
The Conversation Plan is your canonical grounding, not a script; your response MUST be dictated by what the Agent just said.

# Persona Description

{{ persona.description }}
This persona behaves in the following ways:
{% for b in persona.behaviors %}
## {{ b.name | render_string_filter}}
{{ b.description | render_string_filter }}

Instructions:
{{ b.get_behavior_instructions_str() | render_string_filter }}
{% endfor %}
# Conversation Plan

{{ conversation_plan }}

# Conversation History

{{ conversation_history }}
""".strip()


def is_valid_user_simulator_template(
    template_str: str, required_params: list[str]
) -> bool:
  """Checks if the given template_str is a valid jinja template."""
  from jinja2 import exceptions
  from jinja2 import meta
  from jinja2 import StrictUndefined
  from jinja2.sandbox import SandboxedEnvironment

  # StrictUndefined allows us to check for all the given params.
  env = SandboxedEnvironment(undefined=StrictUndefined)
  try:
    # Check syntax of template
    template = env.parse(template_str)

    # Find all variables the template expects
    undeclared_variables = meta.find_undeclared_variables(template)

    # Check parameters in template
    missing_required = [
        v for v in required_params if v not in undeclared_variables
    ]

    return not (missing_required)

  except (
      exceptions.TemplateSyntaxError,
      exceptions.UndefinedError,
  ) as _:
    return False


def _get_user_simulator_instructions_template(
    custom_instructions: Optional[str] = None,
    user_persona: Optional[UserPersona] = None,
) -> str:
  """Returns the appropriate instruction template for the user simulator."""
  if custom_instructions is None and user_persona is None:
    return _DEFAULT_USER_SIMULATOR_INSTRUCTIONS_TEMPLATE

  if custom_instructions is None and user_persona is not None:
    return _USER_SIMULATOR_INSTRUCTIONS_WITH_PERSONA_TEMPLATE

  if custom_instructions is not None and user_persona is None:
    return custom_instructions

  if custom_instructions is not None and user_persona is not None:
    if not is_valid_user_simulator_template(
        custom_instructions,
        required_params=[
            "stop_signal",
            "conversation_plan",
            "conversation_history",
            "persona",
        ],
    ):
      raise ValueError(
          textwrap.dedent(
              """Custom instructions using personas must contain the following formatting placeholders following Jinja syntax:
                * {{ stop_signal }} : text to be generated when the user simulator decides that the
                  conversation is over.
                * {{ conversation_plan }} : the overall plan for the conversation that the user
                  simulator must follow.
                * {{ conversation_history }} : the conversation between the user and the agent so far.
                * {{ persona }} : UserPersona for the simulator to use.
              """
          )
      )

    return custom_instructions


def get_llm_backed_user_simulator_prompt(
    conversation_plan: str,
    conversation_history: str,
    stop_signal: str,
    custom_instructions: Optional[str] = None,
    user_persona: Optional[UserPersona] = None,
):
  """Formats the prompt for the llm-backed user simulator"""
  from jinja2 import DictLoader
  from jinja2 import pass_context
  from jinja2 import Template
  from jinja2.sandbox import SandboxedEnvironment

  templates = {
      "user_instructions": _get_user_simulator_instructions_template(
          custom_instructions=custom_instructions,
          user_persona=user_persona,
      ),
  }
  template_env = SandboxedEnvironment(loader=DictLoader(templates))

  @pass_context
  def _render_string_filter(context, template_string):
    if not template_string:
      return ""
    return Template(template_string).render(context)

  template_env.filters["render_string_filter"] = _render_string_filter

  template_parameters = {
      "stop_signal": stop_signal,
      "conversation_plan": conversation_plan,
      "conversation_history": conversation_history,
  }
  if user_persona is not None:
    template_parameters["persona"] = user_persona

  return template_env.get_template("user_instructions").render(
      template_parameters
  )



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import List
from typing import Optional

from google.genai import types
from typing_extensions import override

from ..agents.callback_context import CallbackContext
from ..agents.readonly_context import ReadonlyContext
from ..models.llm_request import LlmRequest
from .base_planner import BasePlanner

logger = logging.getLogger('google_adk.' + __name__)


class BuiltInPlanner(BasePlanner):
  """The built-in planner that uses model's built-in thinking features.

  Attributes:
      thinking_config: Config for model built-in thinking features. An error
        will be returned if this field is set for models that don't support
        thinking.
  """

  thinking_config: types.ThinkingConfig
  """
  Config for model built-in thinking features. An error will be returned if this
  field is set for models that don't support thinking.
  """

  def __init__(self, *, thinking_config: types.ThinkingConfig):
    """Initializes the built-in planner.

    Args:
      thinking_config: Config for model built-in thinking features. An error
        will be returned if this field is set for models that don't support
        thinking.
    """
    self.thinking_config = thinking_config

  def apply_thinking_config(self, llm_request: LlmRequest) -> None:
    """Applies the thinking config to the LLM request.

    Args:
      llm_request: The LLM request to apply the thinking config to.
    """
    if self.thinking_config:
      llm_request.config = llm_request.config or types.GenerateContentConfig()
      if llm_request.config.thinking_config:
        logger.debug(
            'Overwriting `thinking_config` from `generate_content_config` with '
            'the one provided by the `BuiltInPlanner`.'
        )
      llm_request.config.thinking_config = self.thinking_config

  @override
  def build_planning_instruction(
      self,
      readonly_context: ReadonlyContext,
      llm_request: LlmRequest,
  ) -> Optional[str]:
    return

  @override
  def process_planning_response(
      self,
      callback_context: CallbackContext,
      response_parts: List[types.Part],
  ) -> Optional[List[types.Part]]:
    return



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pydantic models for ADK recordings."""

from __future__ import annotations

from typing import Optional

from google.genai import types
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field

from ...models.llm_request import LlmRequest
from ...models.llm_response import LlmResponse


class LlmRecording(BaseModel):
  """Paired LLM request and response."""

  model_config = ConfigDict(
      extra="forbid",
  )

  llm_request: Optional[LlmRequest] = None
  """Required. The LLM request."""

  llm_responses: Optional[list[LlmResponse]] = None
  """Required. The list of LLM responses."""


class ToolRecording(BaseModel):
  """Paired tool call and response."""

  model_config = ConfigDict(
      extra="forbid",
  )

  tool_call: Optional[types.FunctionCall] = None
  """Required. The tool call."""

  tool_response: Optional[types.FunctionResponse] = None
  """Required. The tool response."""


class Recording(BaseModel):
  """Single interaction recording, ordered by request timestamp."""

  model_config = ConfigDict(
      extra="forbid",
  )

  user_message_index: int
  """Index of the user message this recording belongs to (0-based)."""

  agent_name: str
  """Name of the agent."""

  # oneof fields - start
  llm_recording: Optional[LlmRecording] = None
  """LLM request-response pair."""

  tool_recording: Optional[ToolRecording] = None
  """Tool call-response pair."""
  # oneof fields - end


class Recordings(BaseModel):
  """All recordings in chronological order."""

  model_config = ConfigDict(
      extra="forbid",
  )

  recordings: list[Recording] = Field(default_factory=list)
  """Chronological list of all recordings."""



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provides data for the agent."""

from __future__ import annotations

from typing import Any
from typing import TYPE_CHECKING

from typing_extensions import override

from ..tool_context import ToolContext
from .base_retrieval_tool import BaseRetrievalTool

if TYPE_CHECKING:
  from llama_index.core.base.base_retriever import BaseRetriever


class LlamaIndexRetrieval(BaseRetrievalTool):

  def __init__(self, *, name: str, description: str, retriever: BaseRetriever):
    super().__init__(name=name, description=description)
    self.retriever = retriever

  @override
  async def run_async(
      self, *, args: dict[str, Any], tool_context: ToolContext
  ) -> Any:
    return self.retriever.retrieve(args['query'])[0].text



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Any
from typing import AsyncGenerator
from typing import Optional

from google.genai import types

from ..features import FeatureName
from ..features import is_feature_enabled
from ..models.llm_response import LlmResponse


class StreamingResponseAggregator:
  """Aggregates partial streaming responses.

  It aggregates content from partial responses, and generates LlmResponses for
  individual (partial) model responses, as well as for aggregated content.
  """

  def __init__(self) -> None:
    self._text = ''
    self._thought_text = ''
    self._usage_metadata = None
    self._grounding_metadata: Optional[types.GroundingMetadata] = None
    self._citation_metadata: Optional[types.CitationMetadata] = None
    self._response = None

    # For progressive SSE streaming mode: accumulate parts in order
    self._parts_sequence: list[types.Part] = []
    self._current_text_buffer: str = ''
    self._current_text_is_thought: Optional[bool] = None
    self._finish_reason: Optional[types.FinishReason] = None

    # For streaming function call arguments
    self._current_fc_name: Optional[str] = None
    self._current_fc_args: dict[str, Any] = {}
    self._current_fc_id: Optional[str] = None
    self._current_thought_signature: Optional[bytes] = None

  def _flush_text_buffer_to_sequence(self) -> None:
    """Flush current text buffer to parts sequence.

    This helper is used in progressive SSE mode to maintain part ordering.
    It only merges consecutive text parts of the same type (thought or regular).
    """
    if self._current_text_buffer:
      if self._current_text_is_thought:
        self._parts_sequence.append(
            types.Part(text=self._current_text_buffer, thought=True)
        )
      else:
        self._parts_sequence.append(
            types.Part.from_text(text=self._current_text_buffer)
        )
      self._current_text_buffer = ''
      self._current_text_is_thought = None

  def _get_value_from_partial_arg(
      self, partial_arg: types.PartialArg, json_path: str
  ) -> tuple[Any, bool]:
    """Extract value from a partial argument.

    Args:
      partial_arg: The partial argument object
      json_path: JSONPath for this argument

    Returns:
      Tuple of (value, has_value) where has_value indicates if a value exists
    """
    value: Any = None
    has_value = False

    if partial_arg.string_value is not None:
      # For streaming strings, append chunks to existing value
      string_chunk = partial_arg.string_value
      has_value = True

      # Get current value for this path (if any)
      path_without_prefix = (
          json_path[2:] if json_path.startswith('$.') else json_path
      )
      path_parts = path_without_prefix.split('.')

      # Try to get existing value
      existing_value: Any = self._current_fc_args
      for part in path_parts:
        if isinstance(existing_value, dict) and part in existing_value:
          existing_value = existing_value[part]
        else:
          break

      # Append to existing string or set new value
      if isinstance(existing_value, str):
        value = existing_value + string_chunk
      else:
        value = string_chunk

    elif partial_arg.number_value is not None:
      value = partial_arg.number_value
      has_value = True
    elif partial_arg.bool_value is not None:
      value = partial_arg.bool_value
      has_value = True
    elif partial_arg.null_value is not None:
      value = None
      has_value = True

    return value, has_value

  def _set_value_by_json_path(self, json_path: str, value: Any) -> None:
    """Set a value in _current_fc_args using JSONPath notation.

    Args:
      json_path: JSONPath string like "$.location" or "$.location.latitude"
      value: The value to set
    """
    # Remove leading "$." from jsonPath
    if json_path.startswith('$.'):
      path = json_path[2:]
    else:
      path = json_path

    # Split path into components
    path_parts = path.split('.')

    # Navigate to the correct location and set the value
    current = self._current_fc_args
    for part in path_parts[:-1]:
      if part not in current:
        current[part] = {}
      current = current[part]

    # Set the final value
    current[path_parts[-1]] = value

  def _flush_function_call_to_sequence(self) -> None:
    """Flush current function call to parts sequence.

    This creates a complete FunctionCall part from accumulated partial args.
    """
    if self._current_fc_name:
      # Create function call part with accumulated args
      fc_part = types.Part.from_function_call(
          name=self._current_fc_name,
          args=self._current_fc_args.copy(),
      )

      # Set the ID if provided (directly on the function_call object)
      if self._current_fc_id and fc_part.function_call:
        fc_part.function_call.id = self._current_fc_id

      # Set thought_signature if provided (on the Part, not FunctionCall)
      if self._current_thought_signature:
        fc_part.thought_signature = self._current_thought_signature

      self._parts_sequence.append(fc_part)

      # Reset FC state
      self._current_fc_name = None
      self._current_fc_args = {}
      self._current_fc_id = None
      self._current_thought_signature = None

  def _process_streaming_function_call(self, fc: types.FunctionCall) -> None:
    """Process a streaming function call with partialArgs.

    Args:
      fc: The function call object with partial_args
    """
    # Save function name if present (first chunk)
    if fc.name:
      self._current_fc_name = fc.name
    if fc.id:
      self._current_fc_id = fc.id

    # Process each partial argument
    for partial_arg in fc.partial_args or []:
      json_path = partial_arg.json_path
      if not json_path:
        continue

      # Extract value from partial arg
      value, has_value = self._get_value_from_partial_arg(
          partial_arg, json_path
      )

      # Set the value using JSONPath (only if a value was provided)
      if has_value:
        self._set_value_by_json_path(json_path, value)

    # Check if function call is complete
    if not fc.will_continue:
      # Function call complete, flush it
      self._flush_text_buffer_to_sequence()
      self._flush_function_call_to_sequence()

  def _process_function_call_part(self, part: types.Part) -> None:
    """Process a function call part (streaming or non-streaming).

    Args:
      part: The part containing a function call
    """
    fc = part.function_call
    if fc is None:
      return

    # Check if this is a streaming FC (has partialArgs or will_continue=True)
    # The first chunk of a streaming function call may have will_continue=True
    # but no partial_args yet, so we need to check both conditions.
    if fc.partial_args or fc.will_continue:
      # Streaming function call arguments

      # Save thought_signature from the part (first chunk should have it)
      if part.thought_signature and not self._current_thought_signature:
        self._current_thought_signature = part.thought_signature
      self._process_streaming_function_call(fc)
    else:
      # Non-streaming function call (standard format with args)
      # Skip empty function calls (used as streaming end markers)
      if fc.name:
        # Flush any buffered text first, then add the FC part
        self._flush_text_buffer_to_sequence()
        self._parts_sequence.append(part)

  async def process_response(
      self, response: types.GenerateContentResponse
  ) -> AsyncGenerator[LlmResponse, None]:
    """Processes a single model response.

    Args:
      response: The response to process.

    Yields:
      The generated LlmResponse(s), for the partial response, and the aggregated
      response if needed.
    """
    # results = []
    self._response = response
    llm_response = LlmResponse.create(response)
    self._usage_metadata = llm_response.usage_metadata
    if llm_response.grounding_metadata:
      self._grounding_metadata = llm_response.grounding_metadata
    if llm_response.citation_metadata:
      self._citation_metadata = llm_response.citation_metadata

    # ========== Progressive SSE Streaming (new feature) ==========
    # Save finish_reason for final aggregation
    if llm_response.finish_reason:
      self._finish_reason = llm_response.finish_reason

    if is_feature_enabled(FeatureName.PROGRESSIVE_SSE_STREAMING):
      # Accumulate parts while preserving their order
      # Only merge consecutive text parts of the same type (thought or regular)
      if llm_response.content and llm_response.content.parts:
        for part in llm_response.content.parts:
          if part.text:
            # Check if we need to flush the current buffer first
            # (when text type changes from thought to regular or vice versa)
            if (
                self._current_text_buffer
                and part.thought != self._current_text_is_thought
            ):
              self._flush_text_buffer_to_sequence()

            # Accumulate text to buffer
            if not self._current_text_buffer:
              self._current_text_is_thought = part.thought
            self._current_text_buffer += part.text
          elif part.function_call:
            # Process function call (handles both streaming Args and
            # non-streaming Args)
            self._process_function_call_part(part)
          else:
            # Other non-text parts (bytes, etc.)
            # Flush any buffered text first, then add the non-text part
            self._flush_text_buffer_to_sequence()
            self._parts_sequence.append(part)

      # Mark ALL intermediate chunks as partial
      llm_response.partial = True
      yield llm_response
      return

    # ========== Non-Progressive SSE Streaming (old behavior) ==========
    if (
        llm_response.content
        and llm_response.content.parts
        and llm_response.content.parts[0].text
    ):
      part0 = llm_response.content.parts[0]
      part_text = part0.text or ''
      if part0.thought:
        self._thought_text += part_text
      else:
        self._text += part_text
      llm_response.partial = True
    elif (self._thought_text or self._text) and (
        not llm_response.content
        or not llm_response.content.parts
        # don't yield the merged text event when receiving audio data
        or not llm_response.content.parts[0].inline_data
    ):
      parts = []
      if self._thought_text:
        parts.append(types.Part(text=self._thought_text, thought=True))
      if self._text:
        parts.append(types.Part.from_text(text=self._text))
      yield LlmResponse(
          content=types.ModelContent(parts=parts),
          usage_metadata=llm_response.usage_metadata,
      )
      self._thought_text = ''
      self._text = ''
    yield llm_response

  def close(self) -> Optional[LlmResponse]:
    """Generate an aggregated response at the end, if needed.

    This should be called after all the model responses are processed.

    Returns:
      The aggregated LlmResponse.
    """
    # ========== Progressive SSE Streaming (new feature) ==========
    if is_feature_enabled(FeatureName.PROGRESSIVE_SSE_STREAMING):
      # Always generate final aggregated response in progressive mode
      if self._response and self._response.candidates:
        # Flush any remaining buffers to complete the sequence
        self._flush_text_buffer_to_sequence()
        self._flush_function_call_to_sequence()

        # Use the parts sequence which preserves original ordering
        final_parts = self._parts_sequence

        if final_parts:
          candidate = self._response.candidates[0]
          finish_reason = self._finish_reason or candidate.finish_reason

          return LlmResponse(
              content=types.ModelContent(parts=final_parts),
              grounding_metadata=self._grounding_metadata,
              citation_metadata=self._citation_metadata,
              error_code=None
              if finish_reason == types.FinishReason.STOP
              else finish_reason,
              error_message=None
              if finish_reason == types.FinishReason.STOP
              else candidate.finish_message,
              usage_metadata=self._usage_metadata,
              finish_reason=finish_reason,
              partial=False,
          )

        return None

    # ========== Non-Progressive SSE Streaming (old behavior) ==========
    if (
        (self._text or self._thought_text)
        and self._response
        and self._response.candidates
    ):
      parts = []
      if self._thought_text:
        parts.append(types.Part(text=self._thought_text, thought=True))
      if self._text:
        parts.append(types.Part.from_text(text=self._text))
      candidate = self._response.candidates[0]
      return LlmResponse(
          content=types.ModelContent(parts=parts),
          grounding_metadata=self._grounding_metadata,
          citation_metadata=self._citation_metadata,
          error_code=None
          if candidate.finish_reason == types.FinishReason.STOP
          else candidate.finish_reason,
          error_message=None
          if candidate.finish_reason == types.FinishReason.STOP
          else candidate.finish_message,
          usage_metadata=self._usage_metadata,
      )

    return None



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import enum


class ToolErrorType(str, enum.Enum):
  """HTTP error types conforming to OpenTelemetry semantics."""

  BAD_REQUEST = 'BAD_REQUEST'
  UNAUTHORIZED = 'UNAUTHORIZED'
  FORBIDDEN = 'FORBIDDEN'
  NOT_FOUND = 'NOT_FOUND'
  REQUEST_TIMEOUT = 'REQUEST_TIMEOUT'
  INTERNAL_SERVER_ERROR = 'INTERNAL_SERVER_ERROR'
  BAD_GATEWAY = 'BAD_GATEWAY'
  SERVICE_UNAVAILABLE = 'SERVICE_UNAVAILABLE'
  GATEWAY_TIMEOUT = 'GATEWAY_TIMEOUT'


class ToolExecutionError(Exception):
  """Represents an error that occurs during the execution of a tool."""

  def __init__(
      self, message: str, error_type: ToolErrorType | str | None = None
  ):
    """Initializes the ToolExecutionError exception.

    Args:
      message (str): A message describing the error.
      error_type (ToolErrorType | str | None): The semantic error type (e.g.,
        ToolErrorType.REQUEST_TIMEOUT or '500'). Used to populate the
        `error.type` span attribute in OpenTelemetry traces.
    """
    self.message = message
    if isinstance(error_type, ToolErrorType):
      self.error_type = error_type.value
    else:
      self.error_type = error_type
    super().__init__(self.message)



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for ADK context management.

This module is for ADK internal use only.
Please do not rely on the implementation details.
"""

from __future__ import annotations

from contextlib import aclosing
import inspect
import typing
from typing import Any
from typing import Callable
from typing import get_args
from typing import get_origin
from typing import Union

# Re-export aclosing for backward compatibility
Aclosing = aclosing


def _is_context_type(annotation: Any) -> bool:
  """Check if an annotation is the Context type.

  This checks if the annotation is exactly Context or a type alias of Context
  (e.g., ToolContext, CallbackContext). Also handles Optional[Context] types.

  Args:
    annotation: The type annotation to check.

  Returns:
    True if the annotation is the Context type, False otherwise.
  """
  from ..agents.context import Context

  if annotation is inspect.Parameter.empty:
    return False

  # Handle Optional[Context] and Union types
  origin = get_origin(annotation)
  if origin is Union:
    args = get_args(annotation)
    return any(
        _is_context_type(arg) for arg in args if not isinstance(arg, type(None))
    )

  # Check if it's exactly the Context type (or an alias like ToolContext)
  return annotation is Context


def find_context_parameter(func: Callable[..., Any]) -> str | None:
  """Find the parameter name that has a Context type annotation.

  This function inspects the signature of a callable and returns the name
  of the first parameter that is annotated with Context or a type alias of
  Context (e.g., ToolContext, CallbackContext).

  Args:
    func: The callable to inspect.

  Returns:
    The parameter name if found, None otherwise.
  """
  if func is None:
    return None
  try:
    signature = inspect.signature(func)
  except (ValueError, TypeError):
    return None
  # Resolve string annotations (e.g., 'Context')
  try:
    type_hints = typing.get_type_hints(func)
  except Exception:
    # get_type_hints can fail for various reasons (e.g., unresolvable forward
    # references). In such cases, we fall back to inspecting the parameter
    # annotations directly.
    type_hints = {}

  for name, param in signature.parameters.items():
    annotation = type_hints.get(name, param.annotation)
    if _is_context_type(annotation):
      return name
  return None



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from functools import cached_property
import json
import logging
import re
from typing import Any
from typing import AsyncGenerator

from google.adk.models.google_llm import Gemini
from google.adk.models.llm_request import LlmRequest
from google.adk.models.llm_response import LlmResponse
from google.adk.utils.variant_utils import GoogleLLMVariant
from google.genai import types
from google.genai.types import Content
from google.genai.types import FunctionDeclaration
from google.genai.types import Part
from pydantic import AliasChoices
from pydantic import BaseModel
from pydantic import Field
from pydantic import ValidationError
from typing_extensions import override

logger = logging.getLogger('google_adk.' + __name__)


class GemmaFunctionCallingMixin:
  """Mixin providing function calling support for Gemma models.

  Gemma models don't have native function calling support, so this mixin
  provides the logic to:
  1. Convert function declarations to system instruction prompts
  2. Convert function call/response parts to text in the conversation
  3. Extract function calls from model text responses
  """

  def _move_function_calls_into_system_instruction(
      self, llm_request: LlmRequest
  ) -> None:
    """Converts function declarations to system instructions for Gemma."""
    # Convert function calls/responses in contents to text
    new_contents: list[Content] = []
    for content_item in llm_request.contents:
      (
          new_parts_for_content,
          has_function_response_part,
          has_function_call_part,
      ) = _convert_content_parts_for_gemma(content_item)

      if has_function_response_part:
        if new_parts_for_content:
          new_contents.append(Content(role='user', parts=new_parts_for_content))
      elif has_function_call_part:
        if new_parts_for_content:
          new_contents.append(
              Content(role='model', parts=new_parts_for_content)
          )
      else:
        new_contents.append(content_item)

    llm_request.contents = new_contents

    if not llm_request.config.tools:
      return

    all_function_declarations: list[FunctionDeclaration] = []
    for tool_item in llm_request.config.tools:
      if isinstance(tool_item, types.Tool) and tool_item.function_declarations:
        all_function_declarations.extend(tool_item.function_declarations)

    if all_function_declarations:
      system_instruction = _build_gemma_function_system_instruction(
          all_function_declarations
      )
      llm_request.append_instructions([system_instruction])

    llm_request.config.tools = []

  def _extract_function_calls_from_response(
      self, llm_response: LlmResponse
  ) -> None:
    """Extracts function calls from Gemma text responses."""
    if llm_response.partial or (llm_response.turn_complete is True):
      return

    if not llm_response.content:
      return

    if not llm_response.content.parts:
      return

    if len(llm_response.content.parts) > 1:
      return

    response_text = llm_response.content.parts[0].text
    if not response_text:
      return

    try:
      json_candidate = None

      markdown_code_block_pattern = re.compile(
          r'```(?:(json|tool_code))?\s*(.*?)\s*```', re.DOTALL
      )
      block_match = markdown_code_block_pattern.search(response_text)

      if block_match:
        json_candidate = block_match.group(2).strip()
      else:
        found, json_text = _get_last_valid_json_substring(response_text)
        if found:
          json_candidate = json_text

      if not json_candidate:
        return

      function_call_parsed = GemmaFunctionCallModel.model_validate_json(
          json_candidate
      )
      function_call = types.FunctionCall(
          name=function_call_parsed.name,
          args=function_call_parsed.parameters,
      )
      function_call_part = Part(function_call=function_call)
      llm_response.content.parts = [function_call_part]
    except (json.JSONDecodeError, ValidationError) as e:
      logger.debug(
          'Error attempting to parse JSON into function call. Leaving as text'
          ' response. %s',
          e,
      )
    except Exception as e:
      logger.warning(
          'Error processing Gemma function call response: %s',
          e,
          exc_info=True,
      )


class GemmaFunctionCallModel(BaseModel):
  """Flexible Pydantic model for parsing inline Gemma function call responses."""

  name: str = Field(validation_alias=AliasChoices('name', 'function'))
  parameters: dict[str, Any] = Field(
      validation_alias=AliasChoices('parameters', 'args')
  )


class Gemma(GemmaFunctionCallingMixin, Gemini):
  """Integration for Gemma models exposed via the Gemini API.

  Only Gemma 3 models are supported at this time. For agentic use cases,
  use of gemma-3-27b-it and gemma-3-12b-it are strongly recommended.

  For full documentation, see: https://ai.google.dev/gemma/docs/core/

  NOTE: Gemma does **NOT** support system instructions. Any system instructions
  will be replaced with an initial *user* prompt in the LLM request. If system
  instructions change over the course of agent execution, the initial content
  **SHOULD** be replaced. Special care is warranted here.
  See:
  https://ai.google.dev/gemma/docs/core/prompt-structure#system-instructions

  NOTE: Gemma's function calling support is limited. It does not have full
  access to the
  same built-in tools as Gemini. It also does not have special API support for
  tools and
  functions. Rather, tools must be passed in via a `user` prompt, and extracted
  from model
  responses based on approximate shape.

  NOTE: Vertex AI API support for Gemma is not currently included. This **ONLY**
  supports
  usage via the Gemini API.
  """

  model: str = (
      'gemma-3-27b-it'  # Others: [gemma-3-1b-it, gemma-3-4b-it, gemma-3-12b-it]
  )

  def __repr__(self) -> str:
    return f'{self.__class__.__name__}(model="{self.model}")'

  @classmethod
  @override
  def supported_models(cls) -> list[str]:
    """Provides the list of supported models.

    Returns:
    A list of supported models.
    """

    return [
        r'gemma-3.*',
    ]

  @cached_property
  def _api_backend(self) -> GoogleLLMVariant:
    return GoogleLLMVariant.GEMINI_API

  @override
  async def _preprocess_request(self, llm_request: LlmRequest) -> None:
    self._move_function_calls_into_system_instruction(llm_request=llm_request)

    if system_instruction := llm_request.config.system_instruction:
      contents = llm_request.contents
      instruction_content = Content(
          role='user', parts=[Part.from_text(text=system_instruction)]
      )

      # NOTE: if history is preserved, we must include the system instructions ONLY once at the beginning
      # of any chain of contents.
      if contents:
        if contents[0] != instruction_content:
          # only prepend if it hasn't already been done
          llm_request.contents = [instruction_content] + contents

      llm_request.config.system_instruction = None

    return await super()._preprocess_request(llm_request)

  @override
  async def generate_content_async(
      self, llm_request: LlmRequest, stream: bool = False
  ) -> AsyncGenerator[LlmResponse, None]:
    """Sends a request to the Gemma model.

    Args:
      llm_request: LlmRequest, the request to send to the Gemini model.
      stream: bool = False, whether to do streaming call.

    Yields:
      LlmResponse: The model response.
    """
    # print(f'{llm_request=}')
    assert llm_request.model.startswith('gemma-'), (
        f'Requesting a non-Gemma model ({llm_request.model}) with the Gemma LLM'
        ' is not supported.'
    )

    async for response in super().generate_content_async(llm_request, stream):
      self._extract_function_calls_from_response(response)
      yield response


def _convert_content_parts_for_gemma(
    content_item: Content,
) -> tuple[list[Part], bool, bool]:
  """Converts function call/response parts within a content item to text parts.

  Args:
    content_item: The original Content item.

  Returns:
    A tuple containing:
      - A list of new Part objects with function calls/responses converted to text.
      - A boolean indicating if any function response parts were found.
      - A boolean indicating if any function call parts were found.
  """
  new_parts: list[Part] = []
  has_function_response_part = False
  has_function_call_part = False

  for part in content_item.parts:
    if func_response := part.function_response:
      has_function_response_part = True
      response_text = (
          f'Invoking tool `{func_response.name}` produced:'
          f' `{json.dumps(func_response.response)}`.'
      )
      new_parts.append(Part.from_text(text=response_text))
    elif func_call := part.function_call:
      has_function_call_part = True
      new_parts.append(
          Part.from_text(text=func_call.model_dump_json(exclude_none=True))
      )
    else:
      new_parts.append(part)
  return new_parts, has_function_response_part, has_function_call_part


def _build_gemma_function_system_instruction(
    function_declarations: list[FunctionDeclaration],
) -> str:
  """Constructs the system instruction string for Gemma function calling."""
  if not function_declarations:
    return ''

  system_instruction_prefix = 'You have access to the following functions:\n['
  instruction_parts = []
  for func in function_declarations:
    instruction_parts.append(func.model_dump_json(exclude_none=True))

  separator = ',\n'
  system_instruction = (
      f'{system_instruction_prefix}{separator.join(instruction_parts)}\n]\n'
  )

  system_instruction += (
      'When you call a function, you MUST respond in the format of: '
      """{"name": function name, "parameters": dictionary of argument name and its value}\n"""
      'When you call a function, you MUST NOT include any other text in the'
      ' response.\n'
  )
  return system_instruction


def _get_last_valid_json_substring(text: str) -> tuple[bool, str | None]:
  """Attempts to find and return the last valid JSON object in a string.

  This function is designed to extract JSON that might be embedded in a larger
  text, potentially with introductory or concluding remarks. It will always choose
  the last block of valid json found within the supplied text (if it exists).

  Args:
    text: The input string to search for JSON objects.

  Returns:
    A tuple:
      - bool: True if a valid JSON substring was found, False otherwise.
      - str | None: The last valid JSON substring found, or None if none was
        found.
  """
  decoder = json.JSONDecoder()
  last_json_str = None
  start_pos = 0
  while start_pos < len(text):
    try:
      first_brace_index = text.index('{', start_pos)
      _, end_index = decoder.raw_decode(text[first_brace_index:])
      last_json_str = text[first_brace_index : first_brace_index + end_index]
      start_pos = first_brace_index + end_index
    except json.JSONDecodeError:
      start_pos = first_brace_index + 1
    except ValueError:
      break

  if last_json_str:
    return True, last_json_str
  return False, None


try:
  from google.adk.models.lite_llm import LiteLlm  # noqa: F401
except ImportError as e:
  logger.debug('LiteLlm not available; Gemma3Ollama will not be defined: %s', e)
  LiteLlm = None

if LiteLlm is not None:

  class Gemma3Ollama(GemmaFunctionCallingMixin, LiteLlm):
    """Integration for Gemma 3 models running locally via Ollama.

    This enables fully local agent workflows using Gemma 3 models.
    Requires Ollama to be running with a Gemma 3 model pulled.

    Example:
      ollama pull gemma3:12b
      model = Gemma3Ollama(model="ollama/gemma3:12b")
    """

    def __init__(self, model: str = 'ollama/gemma3:12b', **kwargs):
      super().__init__(model=model, **kwargs)

    def __repr__(self) -> str:
      return f'{self.__class__.__name__}(model="{self.model}")'

    @classmethod
    @override
    def supported_models(cls) -> list[str]:
      return [
          r'ollama/gemma3.*',
      ]

    @override
    async def generate_content_async(
        self, llm_request: LlmRequest, stream: bool = False
    ) -> AsyncGenerator[LlmResponse, None]:
      """Sends a request to Gemma via Ollama/LiteLLM.

      Args:
        llm_request: LlmRequest, the request to send.
        stream: bool = False, whether to do streaming call.

      Yields:
        LlmResponse: The model response.
      """
      self._move_function_calls_into_system_instruction(llm_request)

      async for response in super().generate_content_async(llm_request, stream):
        self._extract_function_calls_from_response(response)
        yield response



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .auth_credential import AuthCredential
from .auth_credential import AuthCredentialTypes
from .auth_credential import OAuth2Auth
from .auth_handler import AuthHandler
from .auth_schemes import AuthScheme
from .auth_schemes import AuthSchemeType
from .auth_schemes import OpenIdConnectWithConfig
from .auth_tool import AuthConfig
from .base_auth_provider import BaseAuthProvider



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configuration for A2A agents."""

from __future__ import annotations

from typing import Any
from typing import Awaitable
from typing import Callable
from typing import Optional
from typing import Union

from a2a.client.middleware import ClientCallContext
from a2a.server.events import Event as A2AEvent
from a2a.types import Message as A2AMessage
from pydantic import BaseModel

from ...a2a.converters.part_converter import A2APartToGenAIPartConverter
from ...a2a.converters.part_converter import convert_a2a_part_to_genai_part
from ...a2a.converters.to_adk_event import A2AArtifactUpdateToEventConverter
from ...a2a.converters.to_adk_event import A2AMessageToEventConverter
from ...a2a.converters.to_adk_event import A2AStatusUpdateToEventConverter
from ...a2a.converters.to_adk_event import A2ATaskToEventConverter
from ...a2a.converters.to_adk_event import convert_a2a_artifact_update_to_event
from ...a2a.converters.to_adk_event import convert_a2a_message_to_event
from ...a2a.converters.to_adk_event import convert_a2a_status_update_to_event
from ...a2a.converters.to_adk_event import convert_a2a_task_to_event
from ...agents.invocation_context import InvocationContext
from ...events.event import Event


class ParametersConfig(BaseModel):
  """Configuration for the parameters passed to the A2A send_message request."""

  request_metadata: Optional[dict[str, Any]] = None
  client_call_context: Optional[ClientCallContext] = None
  # TODO: Add support for requested_extension and
  # message_send_configuration once they are supported by the A2A client.
  #
  # requested_extension: Optional[list[str]] = None
  # message_send_configuration: Optional[MessageSendConfiguration] = None


class RequestInterceptor(BaseModel):
  """Interceptor for A2A requests."""

  before_request: Optional[
      Callable[
          [InvocationContext, A2AMessage, ParametersConfig],
          Awaitable[tuple[Union[A2AMessage, Event], ParametersConfig]],
      ]
  ] = None
  """Hook executed before the agent starts processing the request.

    Returns an Event if the request should be aborted and the Event
    returned to the caller.
  """

  after_request: Optional[
      Callable[
          [InvocationContext, A2AEvent, Event], Awaitable[Union[Event, None]]
      ]
  ] = None
  """Hook executed after the agent has processed the request.

    Returns None if the event should not be sent to the caller.
  """


class A2aRemoteAgentConfig(BaseModel):
  """Configuration for A2A remote agents."""

  # Converts standard A2A Messages into ADK Event.
  a2a_message_converter: A2AMessageToEventConverter = (
      convert_a2a_message_to_event
  )

  # Converts an A2A Task into an ADK Event.
  a2a_task_converter: A2ATaskToEventConverter = convert_a2a_task_to_event

  # Converts A2A TaskStatusUpdateEvents into ADK Event.
  a2a_status_update_converter: A2AStatusUpdateToEventConverter = (
      convert_a2a_status_update_to_event
  )

  # Converts A2A TaskArtifactUpdateEvents into ADK Event.
  a2a_artifact_update_converter: A2AArtifactUpdateToEventConverter = (
      convert_a2a_artifact_update_to_event
  )

  # A low-level hook that converts individual A2A Message Parts
  # into native ADK/GenAI Part objects.
  # This is utilized internally by the other converters.
  a2a_part_converter: A2APartToGenAIPartConverter = (
      convert_a2a_part_to_genai_part
  )

  request_interceptors: Optional[list[RequestInterceptor]] = None



# Copyright 2026 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from google.genai import types as genai_types
from pydantic import BaseModel
from pydantic import Field
from pydantic import field_validator
from pydantic import model_validator
from pydantic_core import ValidationError

from ...features import experimental
from ...features import FeatureName


@experimental(FeatureName.ENVIRONMENT_SIMULATION)
class InjectedError(BaseModel):
  """An error to be injected into a tool call."""

  injected_http_error_code: int
  """Inject http error code to the tool call. Will present as "error_code"
  in the tool response dict."""

  error_message: str
  """Inject error message to the tool call. Will present as
  "error_message" in the tool response dict."""


@experimental(FeatureName.ENVIRONMENT_SIMULATION)
class InjectionConfig(BaseModel):
  """Injection configuration for a tool."""

  injection_probability: float = 1.0
  """Probability of injecting the injected_value."""

  match_args: Optional[Dict[str, Any]] = None
  """Only apply injection if the request matches the match_args.
  If match_args is not provided, the injection will be applied to all
  requests."""

  injected_latency_seconds: float = Field(default=0.0, le=120.0)
  """Inject latency to the tool call. Please note it may not be accurate if                                                                                                                                                                                             │
  the interceptor is applied as after tool callback."""

  random_seed: Optional[int] = None
  """The random seed to use for this injection."""

  injected_error: Optional[InjectedError] = None
  """The injected error."""

  injected_response: Optional[Dict[str, Any]] = None
  """The injected response."""

  @model_validator(mode="after")
  def check_injected_error_or_response(self) -> Self:
    """Checks that either injected_error or injected_response is set."""
    if bool(self.injected_error) == bool(self.injected_response):
      raise ValueError(
          "Either injected_error or injected_response must be set, but not"
          " both, and not neither."
      )
    return self


@experimental(FeatureName.ENVIRONMENT_SIMULATION)
class MockStrategy(enum.Enum):
  """Mock strategy for a tool."""

  MOCK_STRATEGY_UNSPECIFIED = 0

  MOCK_STRATEGY_TOOL_SPEC = 1
  """Use tool specifications to mock the tool response."""

  MOCK_STRATEGY_TRACING = 2
  """Deprecated, please use MOCK_STRATEGY_TOOL_SPEC with tracing input."""


@experimental(FeatureName.ENVIRONMENT_SIMULATION)
class ToolSimulationConfig(BaseModel):
  """Simulation configuration for a single tool."""

  tool_name: str
  """Name of the tool to be simulated."""

  injection_configs: List[InjectionConfig] = Field(default_factory=list)
  """Injection configuration for the tool. If provided, the tool will be
  injected with the injected_value with the injection_probability first,
  the mock_strategy will be applied if no injection config is hit."""

  mock_strategy_type: MockStrategy = MockStrategy.MOCK_STRATEGY_UNSPECIFIED
  """The mock strategy to use."""

  @model_validator(mode="after")
  def check_mock_strategy_type(self) -> Self:
    """Checks that mock_strategy_type is not UNSPECIFIED if no injections."""
    if (
        not self.injection_configs
        and self.mock_strategy_type == MockStrategy.MOCK_STRATEGY_UNSPECIFIED
    ):
      raise ValueError(
          "If injection_configs is empty, mock_strategy_type cannot be"
          " MOCK_STRATEGY_UNSPECIFIED."
      )
    return self


@experimental(FeatureName.ENVIRONMENT_SIMULATION)
class EnvironmentSimulationConfig(BaseModel):
  """Configuration for EnvironmentSimulation."""

  tool_simulation_configs: List[ToolSimulationConfig] = Field(
      default_factory=list
  )
  """A list of tool simulation configurations."""

  simulation_model: str = Field(default="gemini-2.5-flash")
  """The model to use for internal simulator LLM calls (tool analysis, mock responses)."""

  simulation_model_configuration: genai_types.GenerateContentConfig = Field(
      default_factory=lambda: genai_types.GenerateContentConfig(
          thinking_config=genai_types.ThinkingConfig(
              include_thoughts=False,
              thinking_budget=10240,
          )
      ),
  )
  """The configuration for the internal simulator LLM calls."""

  tracing: Optional[str] = None
  """Tracing data (e.g., a prior agent run trace in JSON string format) to
  provide historical context for mock generation. Passed directly to mock
  strategies alongside environment_data."""

  environment_data: Optional[str] = None
  """Environment-specific data (e.g., a minimal database dump in JSON string
   format). This data is passed directly to mock strategies for contextual
   mock generation."""

  @field_validator("tool_simulation_configs")
  @classmethod
  def check_tool_simulation_configs(cls, v: List[ToolSimulationConfig]):
    """Checks that tool_simulation_configs is not empty."""
    if not v:
      raise ValueError("tool_simulation_configs must be provided.")
    seen_tool_names = set()
    for tool_sim_config in v:
      if tool_sim_config.tool_name in seen_tool_names:
        raise ValueError(
            f"Duplicate tool_name found: {tool_sim_config.tool_name}"
        )
      seen_tool_names.add(tool_sim_config.tool_name)
    return v
